Acls = {
  // The value to use as the key in a node where there's no
  // user for its left or right child.
  noKey: null,    // WARNING: Do not change this, or poorly written code that checks for keys being equal to null will break.
  emptyHash: CryptoUtils.hash('empty'),
  debug: true,
  
  // @param curAclVer is optional, but if set, this function ensures that the returned
  // ACL has a higher or equal version than curAclVer
  getAcl: function(objid, ownerUsername, ownerPublicKey, curAclVer) {
    if(curAclVer) {
      console.log('Looking for ACL w/ version >= ' + curAclVer);
    } else {
      curAclVer = 0;
    }
    
    // Fetch ACL by objid
    var tree = Acls.Helper.getAcl(objid, ownerUsername);
    if(!tree)
      throw new Meteor.Error('cannot-find-acl', 'Could not find ACL for object \'' + objid 
        + '\' owned by ' + ownerUsername);

    var rootNode = Acls.Helper.getRootNode(tree);
    if(!rootNode)
      throw new Meteor.Error('cannot-find-root', 'Could not find root node in ACL for object \'' 
        + objid + '\' owned by ' + ownerUsername);
    
    // Check ACL version
    if(tree.aclVer < curAclVer) {
      throw new Meteor.Error('bad-version', 'Asked for ACL version >= ' + curAclVer
        + ' but got version ' + tree.aclVer);
    }

    // Verify ACL tree hash signature
    var aclHash = Acls.Helper.getAclHash(tree, rootNode.hash);

    //console.log('ownerPublicKey: ' + CryptoUtils.serializePublicKey(ownerPublicKey)); 
    //console.log('ACL tree: ' + stri(tree));
    if(CryptoUtils.verifySig(ownerPublicKey, aclHash, tree.signature) == false) { 
      throw new Meteor.Error('acl-sig-verify', 
        'Signature on ACL tree did not verify. Got ACL hash: \'' + aclHash + '\'');
    }
    Trustore.Hashes.set('acl.' + objid, aclHash, ownerUsername);
    
    return { acl: tree, hash: aclHash, rootNodeHash: rootNode.hash };
  },

  // @param initialUser an initial user to add to the ACL, if any
  // @param initialKey the initial content key (or null to generate one) for the ACL
  // this should be a SymmetricKey as defined in crypto-utils.js
  createAcl: function(objid, ownerUsername, ownerMasterKey, ownerPrivateKey, initialKey, initialUser, initialUserFriendKey) {
    var log = function(msg) { console.log('[Acls.createAcl] ' + msg); };
    var aclVer = 1;
    var contentKey = initialKey;
    var numUsers = 0;
    var leftChildId = null;

    if(initialKey != null) {
      log('Setting initial ACL content key, as provided');//: ' + initialKey);
    } else {
      contentKey = CryptoUtils.genSymKey();
      log('Generating content key randomly for ACL for ' + objid + ': ' 
        + CryptoUtils.serializeSymKey(contentKey));
    }

    // We need to encrypt the content key for the owner
    var ownerEncKey = CryptoUtils.encryptKey(ownerMasterKey, contentKey);
    // We need to encrypt the content key for the root's children (if initialUser is set)
    var encKeys = [ Acls.noKey, Acls.noKey ];
    var aclIndexId = null;
    var leftHash = Acls.emptyHash;

    // If we are also adding a user to the ACL, then we need to encrypt the content key
    // for him in the root
    if(initialUser) {
      numUsers = 1;
      log('Also adding ' + initialUser + ' to the ACL');

      // Store the content key in the root node, encrypted for this initial user under
      // his friend key with the owner
      encKeys[0] = CryptoUtils.encryptKey(initialUserFriendKey, contentKey);
      
      var aclIndex = {
        treeId: null,
        isDeleted: false,
        username: initialUser,
        nodeId: null,
        keyIdx: 0,    // this user will be the left child of the root
      }

      aclIndexId = AclIndexes.insert(aclIndex);
      leftHash = Acls.Helper.getIndexHash(aclIndex);
    }
   
    // Create the root node and insert it in the tree
    var rootNode = {
      ownerEncKey: [ ownerEncKey ],
      encKeys: [ encKeys ],
      // for now, we just allow one user being added when the ACL is created, 
      // so these are null until we add a third user to the ACL and the tree is expanded.
      leftChild: null,
      rightChild: null,
      parent: null,
    };

    // We include the ACL index nodes in the hash as well
    rootNode.hash = Acls.Helper.getNodeHashWithChildren(rootNode, leftHash, Acls.emptyHash);
 
    //log('Inserting root node: ' + EJSON.stringify(rootNode, { indent: true }));
    var rootNodeId = AclNodes.insert(rootNode);
    rootNode._id = rootNodeId;

    var acl = {
      objid: objid,
      aclVer: aclVer,
      owner: ownerUsername,
      rootNode: rootNodeId,
      numUsers: numUsers,
      signature: null,    // we fill it in below
    };

    // Sign the root node's hash and the current ACL (its version, object, owner, etc.)
    var aclHash = Acls.Helper.getAclHash(acl, rootNode.hash);
    var signature = CryptoUtils.signString(ownerPrivateKey, aclHash);
    //acl.hash = aclHash; // we don't need to store this, we can compute and check against sig.
    acl.signature = signature;
    log('Signing ACL hash for objid ' + objid + ': SIG(' + aclHash + ') = ' + acl.signature);

    //log('Inserting ACL: ' + EJSON.stringify(acl, { indent: true }));
    var treeId = AclTrees.insert(acl);

    // Update the AclIndex with the tree and the root node ID, if we inserted an initial user
    if(aclIndexId) {
      assert(aclIndexId);
      AclIndexes.update({ _id: aclIndexId }, { $set: {
        treeId: treeId, nodeId: rootNodeId
      }});
    }

    // Add the new trusted hash to trustore too
    Trustore.Hashes.set('acl.' + objid, aclHash, ownerUsername);

    Acls.Helper.verifyTree(rootNode, null, 0);

    //console.log('ACL tree for ' + objid + ' owned by ' + ownerUsername + ': ' + stri(acl));
    //console.log('ACL tree root node for ' + objid + ' owned by ' + ownerUsername + ': ' + stri(rootNode));

    return { treeId: treeId, tree: acl, hash: aclHash, contentKey: contentKey };
  },

  getMembers: function(treeId, aclRootNodeHash) {
    // FIXME: SECURITY: Seems like the server can also lie to us about the number
    // of users, refusing to present some users. But we can fix that by including
    // the number of users in the ACL hash and checking for it here.
    var members = AclIndexes.find({ treeId: treeId, isDeleted: false }).fetch();

    for(var i = 0; i < members.length; i++)
      Acls.Helper.verifyPath(treeId, members[i], members[i].username, aclRootNodeHash);
    
    //console.log('members: ' + stri(members));

    return members;
  },

  verifyAcl: function(objid, ownerUsername, trustedHash) {
    // Fetch ACL by objid
    var tree = Acls.Helper.getAcl(objid, ownerUsername);
    //console.log('[Acls.verifyAcl] Found ACL tree for objid ' + objid + ' owned by ' 
    //  + ownerUsername + ': ' + stri(tree));
    //console.log('trustedHash.verify: ' + trustedHash);

    var rootNode = Acls.Helper.getRootNode(tree);
    //console.log('rootNode: ' + stri(rootNode));

    var hashes = Acls.Helper.getChildrenHashes(rootNode).hashes;
    var leftChildHash = hashes[0];
    var rightChildHash = hashes[1];

    // Hash the root node's data
    // Verify root node hash matches expected one
    var rootNodeHash = Acls.Helper.getNodeHashWithChildren(rootNode, leftChildHash, rightChildHash);
    assert(rootNodeHash != null);
    if(rootNodeHash != rootNode.hash) {
      throw new Meteor.Error('root-hash-verify', 
        'Server is lying about root node hash on ACL. Expected \'' 
        + rootNodeHash + '\', but got \'' + rootNode.hash + '\'');
    }

    // Verify tree hash matches expected one
    var aclHash = Acls.Helper.getAclHash(tree, rootNode.hash);
    if(aclHash != trustedHash) {
      throw new Meteor.Error('acl-hash-verify', 
        'Hash on ACL tree is different than trusted hash \'' + trustedHash 
        + '\'. Got \'' + aclHash + '\'');
    }

    return { tree: tree, rootNode: rootNode, rootNodeHash: rootNodeHash,
      leftNodeHash: leftChildHash, rightNodeHash: rightChildHash,
      aclHash: aclHash, };
  },

  // The added user gets access to all past keys!
  //
  // @param theirUsername person whom you are giving access to
  //
  // @return should return the new ACL hash
  // FIXME: Concurrency issues in Acls.addToAcl
  addToAcl: function(objid, ownerUsername, ownerPrivateKey, ownerMasterKey, theirUsername, trustedHash) {
    var log = function(msg) { console.log('[Acls.addToAcl] ' + msg); };
    //log('trustedHash: ' + trustedHash);
    
    var acl = Acls.verifyAcl(objid, ownerUsername, trustedHash);
    var tree = acl.tree;
    var rootNode = acl.rootNode; 
    var rootNodeHash = acl.rootNodeHash;
    var leftHash = acl.leftNodeHash;
    var rightHash = acl.rightNodeHash;
    var aclHash = acl.aclHash;

    Acls.Helper.verifyTree(rootNode, null, 0);
    
    // NOTE: We do not need to verify the signature here, because we already
    // have a trusted hash. We verify the signature when cookies are cleared
    // and we have to (blindly) trust hashes.

    // Case 1: User was on the ACL before
    // 
    // Check if the person was on the ACL before. If that's the case, then must
    // reencrypt along his path and I'm done. 
    //
    // I can simply get the leaf node's key and encrypt it for the new user.
    // The ancestor nodes all have their keys set up right.
    var aclIndex = Acls.Helper.wasOnAcl(tree._id, theirUsername);
    var friendKey = Friends.getFriendKey(ownerUsername, ownerMasterKey, theirUsername);
    var path = null;
    if(aclIndex) {
      assert.equal(aclIndex.isDeleted, true);
      // Verify tree from AclIndex up to rootNode.hash and get sibling hashes
      path = Acls.Helper.verifyPath(tree._id, aclIndex, theirUsername, rootNodeHash);
    } else {
      log('Must create new AclIndex for ' + theirUsername);

      // Case 2: User needs to be added to the ACL tree for the first time
      // 
      // Case 2.1: There's not enough space in the ACL tree, need to extend it
      // by creating a new root and adding the new user to its right subtree.
      var startInfo = null;
      var numUsersBin = Number(tree.numUsers).toString(2);

      if(!Acls.Helper.hasSpace(tree)) {
        var newTree = Acls.Helper.extendTree(tree, rootNode, ownerUsername, ownerMasterKey);
        rootNode = newTree.newRoot;
        leftHash = newTree.leftSubtreeHash;
        // When extended, the right subtree is empty so there's no verification work
        // needed really,
        startInfo = Acls.Helper.rootInsertionPoint(rootNode, numUsersBin, newTree.rightChildKey);
      } else {
        // First we need to find the node in the right subtree where we start creating
        // the path to the new user's AclIndex leaf.
        //
        // We can use a neat binary tree trick, where the tree's leaf number in binary, gives you
        // the direction (left/right) you need to take at each node to reach it.
        // As we follow the directions at some point the left/right child pointer will be null.
        // That's where we start inserting our path.
        // 
        // We are calling this after previously extending the tree (i.e. on a big enough tree).
        // We need to verify and commit to the nodes received from the server.
        startInfo = Acls.Helper.findInsertionPoint(rootNode, ownerMasterKey, numUsersBin);
      }

      // Need to create a large enough path that ends in an aclIndex and has the node keys
      // encrypted under the owner's key.
      path = Acls.Helper.createPath(tree, startInfo, ownerMasterKey, theirUsername);

      // Update the number of users in the tree
      tree.numUsers++;
    }

    // Now we know the leaf we need to insert at, and we just need 
    // to update the path with the new user
   
    // Add the new user to the path
    path = Acls.Helper.addUserToPath(path, ownerMasterKey, theirUsername, friendKey);

    // Recompute hashes up to rootNode using the sibling hashes and update
    rootNodeHash = Acls.Helper.updatePathHashes(path);

    //log('Root node after adding ' + theirUsername + ': ' + stri(path.nodes[path.nodes.length - 1]));

    // Recompute the ACL hash, sign it, and update the tree in the DB
    trustedHash = Acls.Helper.getAclHash(tree, rootNodeHash);
    tree.signature = CryptoUtils.signString(ownerPrivateKey, trustedHash);
    assert(tree._id);
    //AclTrees.update(tree._id, tree);
    AclTrees.update(tree._id, {$set: {objid: tree.objid, aclVer: tree.aclVer, owner: tree.owner, rootNode: tree.rootNode, numUsers: tree.numUsers, signature: tree.signature}});
                    // Because client doesn't seem to be allowed to replace things in databases.

    Trustore.Hashes.set('acl.' + objid, trustedHash, ownerUsername);

    return trustedHash;
  },

  removeFromAcl: function(objid, ownerUsername, ownerPrivateKey, ownerMasterKey, theirUsername, trustedHash) {
    console.log('[Acls.removeFromAcl] verifying path against trusted ACL hash: ' + trustedHash);
    var acl = Acls.verifyAcl(objid, ownerUsername, trustedHash);

    // Fetch ACL by objid
    var tree = acl.tree;
    var rootNode = acl.rootNode; 
    var rootNodeHash = acl.rootNodeHash;
    var aclHash = acl.aclHash;
 
    Acls.Helper.verifyTree(rootNode, null, 0);

    var aclIndex = Acls.Helper.wasOnAcl(tree._id, theirUsername);
    assert(aclIndex.isDeleted == false);

    // Verify ACL up to the root's trusted hash
    var path = Acls.Helper.verifyPath(tree._id, aclIndex, theirUsername, rootNodeHash);

    // Modify path keys to cryptographically remove the user from the ACL
    path = Acls.Helper.removeUserFromPath(path, ownerUsername, ownerMasterKey, theirUsername);
    
    // Commit path back
    var newRootHash = Acls.Helper.updatePathHashes(path);

    // Return new trusted hash
    tree.aclVer = tree.aclVer + 1;
    var newTrustedHash = Acls.Helper.getAclHash(tree, newRootHash);
    var signature = CryptoUtils.signString(ownerPrivateKey, newTrustedHash);
    tree.signature = signature;
    
    // Update the ACL tree
    assert(tree._id);
    //AclTrees.update(tree._id, tree);
    AclTrees.update(tree._id, {$set: {objid: tree.objid, aclVer: tree.aclVer, owner: tree.owner, rootNode: tree.rootNode, numUsers: tree.numUsers, signature: tree.signature}});

    // Add the new trusted hash to trustore too
    Trustore.Hashes.set('acl.' + objid, newTrustedHash, ownerUsername);

    return newTrustedHash;
  },

  getContentKeys: function(objid, ownerUsername, ownerPublicKey, 
    theirUsername, friendKey, trustedHash) 
  {
    check(trustedHash, String);
    assert(friendKey.length > 0);
    var log = function(msg) { console.log('[Acls.getContentKeys] ' + msg); };
    //log('Getting content keys for acl.' + objid + ' as ' + theirUsername);
    var acl = Acls.verifyAcl(objid, ownerUsername, trustedHash);
    var tree = acl.tree;
    var rootNode = acl.rootNode; 
    var rootNodeHash = acl.rootNodeHash;
    
    var nodeKey = friendKey;
    var contentKeys = [];

    Acls.Helper.verifyTree(rootNode, null, 0);

    // If the owner is querying we just have to decrypt the root 
    if(ownerUsername == theirUsername) {
      //log('For owner: ' + ownerUsername);
      var node = rootNode;
      // the friendKey is a masterKey when the owner is querying the ACL
      
      for(var i = 0; i < node.ownerEncKey.length; i++) {
        contentKeys.push(CryptoUtils.decryptKey(nodeKey, node.ownerEncKey[i]));
      }
    } else {
      //log('For user: ' + theirUsername);
      var aclIndex =  Acls.Helper.wasOnAcl(tree._id, theirUsername);
      if(aclIndex == null)
        throw new Meteor.Error('not-on-acl', 'User ' + theirUsername + ' is not on ACL for \'' 
          + objid + '\'');
      var path = Acls.Helper.verifyPath(tree._id, aclIndex, theirUsername, rootNodeHash);
      var node = path.nodes[1];
      var len = path.nodes.length; 
      var childIdx = path.siblings[0].idx == 0 ? 1 : 0;
      //log(theirUsername + ' AclIndex is a ' + (childIdx == 1 ? 'right' : 'left') + ' child');
      var idx = 1;

      // Go up through the path, decrypting the keys
      while(!Acls.Helper.isRootNode(node)) {
        //log('Decrypting node: ' + stri(node));
        nodeKey = CryptoUtils.decryptKey(nodeKey, node.encKeys[childIdx]);
        //log('nodeKey: ' + CryptoUtils.serializeSymKey(nodeKey));
        childIdx = path.siblings[idx].idx == 0 ? 1 : 0;
        //console.log('Node ' + node._id + ' is a ' + (childIdx ? 'right' : 'left') + ' child of ' + node.parent);
        node = Acls.Helper.getNode(node.parent);
        idx = idx + 1;
      }

      //log('Decrypting keys from root node: ' + stri(node));
      for(var i = 0; i < node.encKeys.length; i++) {
        var keyCtxt = node.encKeys[i][childIdx];
        //log('Decrypting root node\'s ' + (childIdx ? 'right' : 'left')
        //  + ' encKey #' + i + ' with root key: '
        //  + CryptoUtils.serializeSymKey(nodeKey));
        //log('Key ctext: ' + stri(keyCtxt));
        contentKeys.push(CryptoUtils.decryptKey(nodeKey, keyCtxt));
      }
    }

    return contentKeys;
  },

  Helper: {
    getAcl: function(objid, ownerUsername) {
      var acl = AclTrees.findOne({ objid: objid, owner: ownerUsername });
      // Make sure the server's giving us the right ACL
      assert(!acl || acl.objid == objid);
      return acl;
      //aclVer: { $gte: curAclVer } });
    },

    getAclHash: function(tree, rootNodeHash) {
      check(tree.objid, String);
      check(tree.aclVer, Number);
      check(tree.owner, String);
      check(tree.numUsers, Number);
      check(rootNodeHash, String);

      var str = 'acl|' + tree.objid + '|' + tree.aclVer + '|' 
        + tree.numUsers + '|' + tree.owner + '|' + rootNodeHash;
      var h = CryptoUtils.hash(str);

      //console.log('Hashing ACL \'' + str + '\': ' + h);
      return h;
    },

    // @return the hash of the node or Acls.emtpyHash if rootNode == null
    getNodeHash: function(rootNode) {
      if(rootNode == null)
        return Acls.emptyHash;

      var hashes = Acls.Helper.getChildrenHashes(rootNode).hashes;
      var leftChildHash = hashes[0];
      var rightChildHash = hashes[1];

      // Hash the root node's data
      var rootNodeHash = Acls.Helper.getNodeHashWithChildren(rootNode, leftChildHash, rightChildHash);
      //var rootNodeHash = CryptoUtils.hash(rootData);
      //console.log('leftChildHash: ' + leftChildHash);
      //console.log('emptyhash: ' + Acls.emptyHash);
      //console.log('[Acls.getNodeHash] Hashing node ' + rootNode._id + ' w/ left hash '
      //  + leftChildHash + ' and right hash ' + rightChildHash + ' -> ' + rootNodeHash);

      return rootNodeHash;
    },

    getNodeHashWithChildren: function(rootNode, leftChildHash, rightChildHash) {
      // We need to hash the node's keys, so we 
      // just call EJSON.stringify on the array of encrypted content keys, just so
      // we can easily deal with both cases when encKeys is an array or an array of arrays
      var encKeysStr = EJSON.stringify(rootNode.encKeys, { canonical: true });
      var ownerEncKeyStr = EJSON.stringify(rootNode.ownerEncKey, { canonical: true });

      var nodeData = ownerEncKeyStr + '|' + encKeysStr + '|' + leftChildHash + '|' + rightChildHash;

      return CryptoUtils.hash(nodeData);
    },

    verifyChildren: function(parentNode, leftChildHash, rightChildHash) {
      var parentHash = Acls.Helper.getNodeHashWithChildren(
        parentNode, leftChildHash, rightChildHash);

      return parentHash == parentNode.hash;
    },

    getRootNode: function(tree) {
      return AclNodes.findOne({ _id: tree.rootNode });
    },

    hasSpace: function(tree) {
      // check that the number of users on the tree is not a power of two
      // and is not 0 or 1
      return tree.numUsers == 0 || tree.numUsers == 1 || (tree.numUsers & (tree.numUsers - 1)) != 0;
    },

    // @return the node itself or null if id == null
    getNode: function(id) {
      if(id == null)
        return null;

      return AclNodes.findOne({ _id: id });
    },

    // FIXME: Not concurrent. Even worse, it leaves the ACL tree incomplete
    // when it returns, so the caller is responsible for filling the rest in
    //
    // @param tree the tree to extend (verified & trusted)
    // @param rootNode the tree's root node (verified & trusted)
    //
    // @return the new root node with its ID, the left subtree's hash (which we need
    // the caller to be aware to when hasing & signing the new ACL tree) and its 
    // right child key, which the caller uses to "tie in" the new added user 
    // to the tree
    // 
    // return {
    //   newRoot: newRootNode,
    //   newRootId: newRootId,
    //   leftSubtreeHash: newLeftChild.hash,
    //   rightChildKey: newRightChildKey,
    // };
    extendTree: function(tree, rootNode, ownerUsername, ownerMasterKey) {
      var log = function(msg) { console.log('[extendTree] ' + msg); };
      log('Extending tree...');
      
      // Step 1: Create a new left child for the new root that replaces
      // the old root and is linked with the old left & right children of the old root
      
      // We need to check the hashes on the ACL indexes as well, otherwise the
      // server can lie to us and make us reencrypt keys for revoked users.
      var oldHashes = Acls.Helper.getChildrenHashes(rootNode);
      var oldChildren = oldHashes.children;

      var oldLeftChild = oldChildren.nodes[0];
      var oldRightChild = oldChildren.nodes[1];
 
      var oldLeftChildHash = oldHashes.hashes[0];
      var oldRightChildHash = oldHashes.hashes[1];

      // Verify that the old left and right children belong to this tree, since
      // we are going to link them to a new root and we wanna make sure we are
      // linking the right children.
      if(!Acls.Helper.verifyChildren(rootNode, oldLeftChildHash, oldRightChildHash)) {
        throw 'Children hashes did not match parent hash. Something\'s fishy';
      }

      // Get the keys of the old left & right child of the old root
      var oldKeys = Acls.Helper.getChildrenKeys(rootNode, ownerUsername, ownerMasterKey,
        oldLeftChild, oldRightChild);

      // Generate a key for the new root's new left child
      // We'll use this key later to encrypt the new root's content key(s)
      var newLeftChildKey = CryptoUtils.genSymKey();

      // Create the new left child node
      var newLeftChild = {
        ownerEncKey: CryptoUtils.encryptKey(ownerMasterKey, newLeftChildKey), 
        encKeys: [ 
          oldKeys.left == Acls.noKey ? Acls.noKey : CryptoUtils.encryptKey(oldKeys.left, newLeftChildKey),
          oldKeys.right == Acls.noKey ? Acls.noKey : CryptoUtils.encryptKey(oldKeys.right, newLeftChildKey),
        ],
        // This will be the old left child, or null if we had a tree w/ 2 leafs
        leftChild: rootNode.leftChild,
        // This will be the old right child, or null if we had a tree w/ 2 leafs
        rightChild: rootNode.rightChild,
        parent: null,
      };

      // Compute the hash of this new left child node. We have everything in it now
      newLeftChild.hash = Acls.Helper.getNodeHashWithChildren(
        newLeftChild, oldLeftChildHash, oldRightChildHash);

      // Insert the new left child node
      var newLeftChildId = AclNodes.insert(newLeftChild);
      newLeftChild._id = newLeftChildId;

      // If the old root had a left child, then it also had a right child.
      if(oldChildren.type == 'nodes') {
        // Update old left child's parent pointer
        Acls.Helper.updateParent(oldLeftChild._id, newLeftChildId);
        // Update old right child's parent pointer
        Acls.Helper.updateParent(oldRightChild._id, newLeftChildId);
      } else {
        // Update ACL index for the 1st & 2nd user to point to new left child instead of root
        // This call updates both
        Acls.Helper.updateIndex(rootNode._id, newLeftChildId);
      }

      var newRightChildKey = CryptoUtils.genSymKey();

      assert.equal(rootNode.encKeys.length, rootNode.ownerEncKey.length);

      // Step 2: Create the new root and encrypt it the new left child's key and the
      // new right child's key, which this function returns to the caller so it can
      // continue creating the ACL path to the newly added user
      var newRootNode = {
        // NOTE: cannot compute hash yet because caller will need to add new node 
        // to the right subtree. But will have to put a valid hash in as if the right
        // subtree is empty so that createPath can verify it and proceed.
        hash: null,
        ownerEncKey: [],  // we just copy these from the old root below
        encKeys: [],      // we reencrypt these from the old root below
        leftChild: newLeftChildId,
        rightChild: null,   // we don't know the right child until the caller creates it
        parent: null,
      };

      for(var i = 0; i < rootNode.ownerEncKey.length; i++) {
        // FIXME: the plaintexts remain the same, but should we reencrypt?
        newRootNode.ownerEncKey.push(rootNode.ownerEncKey[i]);
        var contentKey = CryptoUtils.decryptKey(ownerMasterKey, rootNode.ownerEncKey[i]);

        var encKeyLeft = CryptoUtils.encryptKey(newLeftChildKey, contentKey);
        var encKeyRight = CryptoUtils.encryptKey(newRightChildKey, contentKey);

        //log('Encrypting keyLeft[' + i + '] =  \'' + CryptoUtils.serializeSymKey(keyLeft) 
        //  + '\' w/ left key \'' + CryptoUtils.serializeSymKey(newLeftChildKey) + '\': ' 
        //  + stri(encKeyLeft));
        //log('Encrypting keyRight[' + i + '] = \'' + CryptoUtils.serializeSymKey(keyRight)
        //  + '\' w/ right key \'' + CryptoUtils.serializeSymKey(newRightChildKey) + '\': ' 
        //  + stri(encKeyRight));

        newRootNode.encKeys.push([encKeyLeft, encKeyRight]);
      }

      newRootNode.hash = Acls.Helper.getNodeHashWithChildren(newRootNode, newLeftChild.hash, Acls.emptyHash);

      // Insert the root node, the caller will update it later with the hash
      var newRootId = AclNodes.insert(newRootNode);
      newRootNode._id = newRootId;
      tree.rootNode = newRootId;
      newLeftChild.parent = newRootId;

      Acls.Helper.updateParent(newLeftChild._id, newRootId);
      //log('New left child inserted: ' + stri(newLeftChild));
      //log('New root inserted: ' + stri(newRootNode));

      // Return new root node and with its ID, the left subtree's hash and 
      // the (not yet created) right child's key
      return { 
        newRoot: newRootNode, 
        newRootId: newRootId,
        leftSubtreeHash: newLeftChild.hash,
        rightChildKey: newRightChildKey,
      };
    },

    getChildren: function(node) {
      var children = { nodes: [ null, null ], type: 'index' };

      // This will imply rightChild is also null
      if(node.leftChild == null) {
        aclIndex = AclIndexes.find({ nodeId: node._id });

        // Only one or two usernames should have that node as a parent
        var count = aclIndex.count();
        assert(count >= 0);
        assert(count <= 2);

        if(count) {
          aclIndex = aclIndex.fetch();
          children.nodes[aclIndex[0].keyIdx] = aclIndex[0];
          if(count == 2) {
            assert.notEqual(aclIndex[0].keyIdx, aclIndex[1].keyIdx);  // should be different children
            children.nodes[aclIndex[1].keyIdx] = aclIndex[1];
          }
        }
        children.type = 'index';
      } else {
        children.nodes[0] = Acls.Helper.getNode(node.leftChild);
        children.nodes[1] = Acls.Helper.getNode(node.rightChild);
        children.type = 'nodes';
      }

      return children;
    },

    getChildrenHashes: function(node) {
      var children = Acls.Helper.getChildren(node);

      //console.log('[Acls.getChildrenHashes] parent ' + stri(node));
      //console.log('[Acls.getChildrenHashes] children ' + stri(children));

      var hashes = { children: children, hashes: [ null, null ] };
      if(children.type == 'nodes') {
        hashes.hashes[0] = children.nodes[0].hash;
        hashes.hashes[1] = children.nodes[1] == null ? Acls.emptyHash : children.nodes[1].hash;
      } else {
        hashes.hashes[0] = Acls.Helper.getIndexHash(children.nodes[0]);
        hashes.hashes[1] = Acls.Helper.getIndexHash(children.nodes[1]);
      }
      
      return hashes;
    },

    // @return if node is a leaf, returns the friend keys of the users underneath it, if
    // it is a non-leaf it returns its left & right child's keys
    getChildrenKeys: function(node, ownerUsername, ownerMasterKey, leftChild, rightChild) {
      var keys = {
        left: Acls.noKey,
        right: Acls.noKey,
      };

      // If this is a leaf node
      if(node.leftChild == null) {
        assert(node.rightChild == null);

        aclIndex = AclIndexes.find({ nodeId: node._id });

        // Only one or two usernames should have that node as a parent
        var count = aclIndex.count();
        assert(count <= 2);

        aclIndex = aclIndex.fetch();
        var users = [ null, null ];
        users[aclIndex[0].keyIdx] = aclIndex[0];
        if(count == 2) {
          assert.notEqual(aclIndex[0].keyIdx, aclIndex[1].keyIdx);  // should be different children
          users[aclIndex[1].keyIdx] = aclIndex[1];
        }
        
        if(users[0].isDeleted == false) {
          keys.left = Friends.getFriendKey(ownerUsername, ownerMasterKey, users[0].username);
        }

        if(count == 2 && users[1].isDeleted == false) {
          keys.right = Friends.getFriendKey(ownerUsername, ownerMasterKey, users[1].username);
        }
      } else {
        keys.left = CryptoUtils.decryptKey(ownerMasterKey, leftChild.ownerEncKey);
        keys.right = CryptoUtils.decryptKey(ownerMasterKey, rightChild.ownerEncKey);
      }

      return keys;
    },

    updateParent: function(nodeId, newParentId) {
      assert(nodeId);
      assert(newParentId);
      AclNodes.update({ _id: nodeId }, { $set: { parent: newParentId }});
    },

    updateIndex: function(nodeId, newNodeId) {
      // This should update both ACL index entries (for left/right usernames associated w/ leaf)
      // Might need to relax this constraint depending on how code evolves

      //console.log('indexes for ' + nodeId + ': ' + EJSON.stringify(
      //  AclIndexes.find({ nodeId: nodeId }).fetch(), { indent: true, canonical: true }));
      assert(nodeId);
      //AclIndexes.update({ nodeId: nodeId, keyIdx: 0 }, { $set: { nodeId: newNodeId }});
      //AclIndexes.update({ nodeId: nodeId, keyIdx: 1 }, { $set: { nodeId: newNodeId }});
      var indexId = AclIndexes.findOne({ nodeId: nodeId, keyIdx: 0 })._id;              // Can only update using ids from client.
      AclIndexes.update(indexId, { $set: { nodeId: newNodeId }});
      indexId = AclIndexes.findOne({ nodeId: nodeId, keyIdx: 1 })._id;
      AclIndexes.update(indexId, { $set: { nodeId: newNodeId }});

      //console.log('updated indexes for ' + newNodeId + ': ' + EJSON.stringify(
      //  AclIndexes.find({ nodeId: newNodeId }).fetch(), { indent: true, canonical: true }));
    },

    // @return the ACL index object if the user is on the ACL (whether deleted or not) 
    // or null otherwise
    wasOnAcl: function(treeId, username) {
      var aclIndex = AclIndexes.findOne({ treeId: treeId, username: username });

      if(aclIndex) {
        //console.log('[Acls.wasOnAcl] Found ACL for ' + username + ': ' + stri(aclIndex));
        return aclIndex;
      } else {
        //console.log('[Acls.wasOnAcl] ' + username + ' was not on the ACL');
        return null;
      }
    },

    getIndexHash: function(aclIndex) {
      if(aclIndex) {
        return CryptoUtils.hash(aclIndex.username + '|' + aclIndex.isDeleted);
      } else {
        return Acls.emptyHash;
      }
    },

    // FIXME: not concurrent
    //
    // Verifies the ACL for the specified username and returns the ACL tree
    // path for that user along with its sibling hashes, so that the path 
    // can be updated securely later.
    verifyPath: function(treeId, aclIndex, theirUsername, rootHash) {
      var nodeId = aclIndex.nodeId;
      var node = Acls.Helper.getNode(nodeId);
      var leftHash = Acls.Helper.getIndexHash(aclIndex);
      var rightHash = Acls.Helper.getIndexHash(AclIndexes.findOne({ 
        treeId: treeId, nodeId: nodeId, username: { $ne: theirUsername } }));

      var siblingHash = rightHash;
      var siblingIdx = 1;
      var siblings = [ ];
      var path = [ aclIndex ];

      if(aclIndex.keyIdx != 0) {
        var tmp = rightHash;
        rightHash = leftHash;
        leftHash = tmp;

        siblingIdx = 0;
      }

      var prevNode = null;
      // Go up through the path, verifying parent node hashes against their children's hashes
      // and recording all the sibling hashes so we can use them when we update the tree
      while(node) {
        // Add our sibling's hash to our list
        siblings.push( { hash: siblingHash, idx: siblingIdx });
        // Add the current node to the path
        path.push(node);

        // Note: it's important that we actually verify the node's contents here!
        if(node.hash != Acls.Helper.getNodeHashWithChildren(node, leftHash, rightHash)) {
          //console.log('[Acls.verifyPath] Node ID ' + node._id + ' did not verify (node.hash = ' 
          //  + node.hash + ')');
          throw new Meteor.Error('path-integrity', 'The ACL path from ' + theirUsername 
            + ' to the root did not verify correctly. Something\'s fishy.');
          break;
        }

        prevNode = node;
        node = Acls.Helper.getNode(node.parent);
        // We might be at the root, so we could be done.
        if(node) {
          var sibling = Acls.Helper.getSibling(node, prevNode);
          // NOTE: We don't need to verify the sibling hash. We just need
          // to make the server commit to it, by including it in our
          // verification. The server can try lying about it but
          // it will get caught when the parent hashes don't match up
          siblingHash = sibling.node.hash;
          siblingIdx = sibling.isLeftChild ? 0 : 1;
          leftHash = sibling.isLeftChild ? sibling.node.hash : prevNode.hash;
          rightHash = sibling.isLeftChild ? prevNode.hash : sibling.node.hash;
        }
      }
      
      assert(prevNode != null);
      assert.equal(path.length, siblings.length + 1);

      // We've verified the root against its children in the loop, now we gotta
      // make sure it's the right (trusted) root!
      if(prevNode.hash != rootHash) {
        //console.log('[Acls.verifyPath] Root hash did not verify');
        throw new Meteor.Error('path-integrity', 'The ACL path from ' + theirUsername 
          + ' to the root did not verify correctly. Something\'s fishy.');
      }

      return {
        siblings: siblings,
        nodes: path,
      }; 
    },

    removeUserFromPath: function(path, ownerUsername, ownerMasterKey, theirUsername) {
      var log = function(msg) { console.log('[Acls.removeUserFromPath] ' + msg); };
      var len = path.nodes.length;
      var nodes = path.nodes;

      var aclIndex = nodes[0];
 
      //log('Sibling of ACL index for ' + aclIndex.username + ': ' + stri(path.siblings[0]));
      var childIdx = path.siblings[0].idx == 0 ? 1 : 0;
      //log('ACL index for ' + aclIndex.username + ': ' + stri(aclIndex));
      assert(aclIndex._id);

      // Mark the user as deleted
      aclIndex.isDeleted = true;

      var siblingHash = path.siblings[0].hash;
      var siblingKey = Acls.Helper.getVerifiedSiblingKey(ownerUsername, ownerMasterKey, aclIndex, siblingHash, 'index');
      var leftChildKey = childIdx == 0 ? Acls.noKey : siblingKey;
      var rightChildKey = childIdx == 1 ? Acls.noKey : siblingKey;
      
      for(var i = 1; i < len - 1; i++) {
        assert(nodes[i]._id);
        var nodeKey = CryptoUtils.genSymKey();

        var encLeftKey = Acls.noKey;
        if(leftChildKey)
          encLeftKey = CryptoUtils.encryptKey(leftChildKey, nodeKey);

        var encRightKey = Acls.noKey;
        if(rightChildKey)
          encRightKey = CryptoUtils.encryptKey(rightChildKey, nodeKey);

        //log('Reencrypting node ID ' + nodes[i]._id + ' with left key \''
        //  + (leftChildKey ? CryptoUtils.serializeSymKey(leftChildKey) : 'none') 
        //  + '\' and right key \''
        //  + (rightChildKey ? CryptoUtils.serializeSymKey(rightChildKey) : 'none') + '\'');

        assert(nodes[i].encKeys.length == 2);
        nodes[i].encKeys[0] = encLeftKey;
        nodes[i].encKeys[1] = encRightKey;
        nodes[i].ownerEncKey = CryptoUtils.encryptKey(ownerMasterKey, nodeKey);

        //log('Getting sibling of ' + stri(nodes[i]));
        siblingHash = path.siblings[i].hash;
        siblingKey = 
          siblingHash == Acls.emptyHash ? Acls.noKey : 
          Acls.Helper.getVerifiedSiblingKey(ownerUsername, ownerMasterKey, nodes[i], siblingHash, 'node');

        leftChildKey = path.siblings[i].idx == 0 ? siblingKey : nodeKey;
        rightChildKey = path.siblings[i].idx == 1 ? siblingKey : nodeKey;
      }

      // The root is special so do it at the very end
      var newContentKey = CryptoUtils.genSymKey();
      var rootNode = nodes[len - 1];

      // DEBUG: check the root is consistent first! NOTE: cannot due this because
      // I don't have the old left & right keys, would need them first.
      var contentKeys = [];
      for(var i = 0; i < rootNode.ownerEncKey.length; i++) {
        var ownerKey = CryptoUtils.decryptKey(ownerMasterKey, rootNode.ownerEncKey[i]);
        //var keyLeft = CryptoUtils.decryptKey(oldLeftChildKey, rootNode.encKeys[i][0]);
        //var keyRight = CryptoUtils.decryptKey(oldRightChildKey, rootNode.encKeys[i][1]);
        //assert.deepEqual(keyLeft, keyRight);
        //assert.deepEqual(keyLeft, ownerKey);
        contentKeys.push(ownerKey);
      }
      
      // Add the new content key to the owner encrypted keys array
      rootNode.ownerEncKey.push(CryptoUtils.encryptKey(ownerMasterKey, newContentKey));
      contentKeys.push(newContentKey);
      rootNode.encKeys = [];

      for(var i = 0; i < rootNode.ownerEncKey.length; i++) {
        var encKeyLeft = Acls.noKey;
        if(leftChildKey != Acls.noKey)
          encKeyLeft = CryptoUtils.encryptKey(leftChildKey, contentKeys[i]);

        var encKeyRight = Acls.noKey;
        if(rightChildKey != Acls.noKey)
          encKeyRight = CryptoUtils.encryptKey(rightChildKey, contentKeys[i]);

        //log('Encrypting keyLeft[' + i + '] =  \'' + CryptoUtils.serializeSymKey(keyLeft) 
        //  + '\' w/ left key \'' + CryptoUtils.serializeSymKey(newLeftChildKey) + '\': ' 
        //  + stri(encKeyLeft));
        //log('Encrypting keyRight[' + i + '] = \'' + CryptoUtils.serializeSymKey(keyRight)
        //  + '\' w/ right key \'' + CryptoUtils.serializeSymKey(newRightChildKey) + '\': ' 
        //  + stri(encKeyRight));

        rootNode.encKeys.push([encKeyLeft, encKeyRight]);
      }

      //log('Root node after removing ' + theirUsername + ': ' + stri(rootNode));

      return path;
    },
    
    addUserToPath: function(path, ownerMasterKey, theirUsername, friendKey) {
      // First, we update the ACL index to indicate the user is now on the ACL
      var aclIndex = path.nodes[0];
      aclIndex.isDeleted = false;
      assert.equal(aclIndex.username, theirUsername);
      
      // We need to know which child we are, so we can store our key
      var childIdx = path.siblings[0].idx == 0 ? 1 : 0;
      var isRoot = Acls.Helper.isRootNode(path.nodes[1]);
      var len = 0;

      if(isRoot) {
        len = path.nodes[1].encKeys.length;
        //console.log('path.nodes[1]: ' + stri(path.nodes[1]));
        //console.log('path.nodes[1].ownerEncKey type: ' + typeof path.nodes[1].ownerEncKey);
        assert.equal(len, path.nodes[1].ownerEncKey.length);
      }

      // If the leaf is the root, then we need to reencrypt all the content keys for 
      // the newly added user
      if(isRoot) {
        // The ownerEncKey field does not need updating when we add someone back.
        //
        // However, we need to be sure to reencrypt all the entries in the encrypted
        // keys array, at the proper index.
        for(var i = 0; i < len; i++) {
          var leafKey = path.nodes[1].ownerEncKey[i];
          var encLeafKey = CryptoUtils.encryptKey(friendKey,
            CryptoUtils.decryptKey(ownerMasterKey, leafKey));
          
          path.nodes[1].encKeys[i][childIdx] = encLeafKey;
        }
      } else {
        var leafKey = path.nodes[1].ownerEncKey;
        var encLeafKey = CryptoUtils.encryptKey(friendKey,
          CryptoUtils.decryptKey(ownerMasterKey, leafKey));

        path.nodes[1].encKeys[childIdx] = encLeafKey;
      }

      return path;
    },

    // FIXME: not concurrent
    // NOTE: modifies the ACL index entry's isDeleted flag to false
    //
    // @return the new root hash
    updatePathHashes: function(path) {
      var log = function(msg) { console.log('[Acls.updatePathHashes] ' + msg); };
      var len = path.nodes.length;
      var nodes = path.nodes;

      // We rehash the ACL index
      var aclIndex = nodes[0];
      var aclIndexHash = Acls.Helper.getIndexHash(aclIndex);
 
      // Next, we update the hashes up the tree and encrypt the leaf node's key
      // under the friend key of the added user.
      // If my sibling is idx 0, I am idx 1
      //log('Sibling of ACL index for ' + aclIndex.username + ': ' + stri(path.siblings[0]));
      var childIdx = path.siblings[0].idx == 0 ? 1 : 0;
      var leftHash = path.siblings[0].idx == 0 ? path.siblings[0].hash : aclIndexHash;
      var rightHash = path.siblings[0].idx == 0 ? aclIndexHash : path.siblings[0].hash;
     
      //log('ACL index for ' + aclIndex.username + ': ' + stri(aclIndex)); 
      assert(aclIndex._id);
      //AclIndexes.update(aclIndex._id, aclIndex);
      AclIndexes.update(aclIndex._id, {$set: {treeId: aclIndex.treeId, nodeId: aclIndex.nodeId, isDeleted: aclIndex.isDeleted, keyIdx: aclIndex.keyIdx, username: aclIndex.username}});

      // We update the parent node of the index (i.e. the leaf node)
      nodes[1].hash = Acls.Helper.getNodeHashWithChildren(nodes[1], leftHash, rightHash);
      assert(nodes[1]._id);
      //AclNodes.update(nodes[1]._id, nodes[1]);
      AclNodes.update(nodes[1]._id, {$set: {ownerEncKey: nodes[1].ownerEncKey, encKeys: nodes[1].encKeys, leftChild: nodes[1].leftChild, rightChild: nodes[1].rightChild, parent: nodes[1].parent, hash: nodes[1].hash}});
      //log('Updating node ID ' + nodes[1]._id + ' with hash ' + nodes[1].hash);
        
      for(var i = 2; i < len; i++) {
        leftHash = path.siblings[i-1].idx == 0 ? path.siblings[i-1].hash : nodes[i-1].hash;
        rightHash = path.siblings[i-1].idx == 0 ? nodes[i-1].hash : path.siblings[i-1].hash;

        nodes[i].hash = Acls.Helper.getNodeHashWithChildren(nodes[i], leftHash, rightHash);
        assert(nodes[i]._id);
        //AclNodes.update(nodes[i]._id, nodes[i]);
        AclNodes.update(nodes[i]._id, {$set: {ownerEncKey: nodes[i].ownerEncKey, encKeys: nodes[i].encKeys, leftChild: nodes[i].leftChild, rightChild: nodes[i].rightChild, parent: nodes[i].parent, hash: nodes[i].hash}});
      //log('Updating node ID ' + nodes[i]._id + ' with hash ' + nodes[i].hash);
      }

      return nodes[len - 1].hash;
    },
    
    getSibling: function(parentNode, node) {
        var isLeftChild = (parentNode.leftChild == node._id);
        //console.log('node ' + node._id + ' with parent ' + parentNode._id 
        //  + ', isLeftChild: ' + isLeftChild);
        if(isLeftChild) {
          if(parentNode.rightChild) {
            var rightChild = Acls.Helper.getNode(parentNode.rightChild);
            //console.log('isLeftChild was true, returning right sibling: rightChild');
            return { node: rightChild, isLeftChild: false };
          } else {
            //console.log('isLeftChild was true, but this node does not have a right sibling');
            return { node: { hash: Acls.emptyHash }, isLeftChild: false };
          }
        } else {
          //console.log('isLeftChild was false, returning left sibling');
          return { node: Acls.Helper.getNode(parentNode.leftChild), isLeftChild: true };
        }
    },

    getVerifiedSiblingKey: function(ownerUsername, ownerMasterKey, node, siblingHash, type) {
      if(type == 'index') {
        var sibAclIndex = AclIndexes.findOne({ nodeId: node.nodeId, username: { $ne: node.username } });
        if(sibAclIndex) {
          var hash = Acls.Helper.getIndexHash(sibAclIndex);
          if(hash != siblingHash) {
            throw new Meteor.Error('integrity-check', 'Server is lying about the ACL index. ' 
              + 'Expected \'' + siblingHash + '\' as the hash, but got \'' + hash + '\'');
          }

          if(sibAclIndex.isDeleted)
            return Acls.noKey;
          else
            return Friends.getFriendKey(ownerUsername, ownerMasterKey, sibAclIndex.username);
        } else {
          if(siblingHash != Acls.emptyHash) {
            throw new Meteor.Error('integrity-check', 'Server is lying about the ACL index. Expected a valid existing sibling ACL index, but got none.');
          }
          return Acls.noKey;
        }
      } else if(type == 'node') {
        assert(node.parent);
        var parentNode = Acls.Helper.getNode(node.parent);
        //console.log('Parent node: ' + stri(parentNode));

        var sib = Acls.Helper.getSibling(parentNode, node);
        var sibNode = sib.node;
        if(sibNode.hash != Acls.emptyHash) {
          var hash = Acls.Helper.getNodeHash(sibNode);
          if(hash != siblingHash) {
            throw new Meteor.Error('integrity-check', 'Server is lying about the ACL index. ' 
              + 'Expected \'' + siblingHash + '\' as the hash, but got \'' + hash + '\'');
          }
        } else {
          if(siblingHash != Acls.emptyHash) {
            throw new Meteor.Error('integrity-check', 'Server is lying about the sibling node. Expected a valid existing sibling node, but got none.');
          }
        }

        return CryptoUtils.decryptKey(ownerMasterKey, sibNode.ownerEncKey);
      } else {
        throw new Meteor.Error('unsupported-sibling-type', 
          'Invalid sibling type. Expected \'node\' or \'index\'');
      }
    },

    isRootNode: function(node) { return node.parent == null; },

    rootInsertionPoint: function(rootNode, leafNoBinStr, rightChildKey) {
      var topPathRev = {
        siblings: [ /* { hash: null, idx: 0 } */ ],
        nodes: [ rootNode ],
      };

      var startInfo = {
        node: rootNode, nodeKey: rightChildKey, childIdx: 1, leafNo: leafNoBinStr, 
        idx: 0, topPathRev: topPathRev };
      //console.log('Root insertion point: ' + stri(startInfo));
      return startInfo;
    },
    
    findInsertionPoint: function(rootNode, ownerMasterKey, leafNoBinStr) {
      var log = function(msg) { console.log('[Acls.findInsertionPoint] ' + msg); };
      var topPathRev = {
        siblings: [ /* { hash: null, idx: 0 } */ ],
        nodes: [ rootNode ],
      };
      var node = rootNode;
      var i = 0;

      var takeLeft;
      //log('Finding insertion point for ACL #' + leafNoBinStr + 'b');
      assert(leafNoBinStr.length > 0);
      // This loop should always break out when it finds an empty subtree or an empty/
      // nonexistent AclIndex.
      // The node where we need to start the path goes unverified by the end of the 
      // loop. We cannot skip verification of this last node, so we delegate it to createPath.
      for(; i < leafNoBinStr.length; i++) {
        takeLeft = (leafNoBinStr[i] == '0');
        //console.log('takeLeft: ' + takeLeft);

        if(takeLeft) {
          if(node.leftChild == null) {
            break;
          } else {
            var leftNode = Acls.Helper.getNode(node.leftChild);
            var rightNodeHash = node.rightChild ? Acls.Helper.getNode(node.rightChild).hash : Acls.emptyHash;

            // NOTE: we hash the node, and take its children's hashes for granted. The left hash
            // however, will be verified in the next iteration.
            if(node.hash != Acls.Helper.getNodeHashWithChildren(node, leftNode.hash, rightNodeHash))
              throw new Meteor.Error('path-integrity', 'The ACL path from the root down towards '
                + 'an insertion point did not verify correctly. Something\'s fishy');

            topPathRev.nodes.push(leftNode);
            topPathRev.siblings.push({ hash: rightNodeHash, idx: 1 });
            node = leftNode;
          }
        } else {
          if(node.rightChild == null) {
            break;
          } else {
            var rightNode = Acls.Helper.getNode(node.rightChild);
            var leftNodeHash = Acls.Helper.getNode(node.leftChild).hash;

            if(node.hash != Acls.Helper.getNodeHashWithChildren(node, leftNodeHash, rightNode.hash))
               throw new Meteor.Error('path-integrity', 'The ACL path from the root down towards '
                 + 'an insertion point did not verify correctly. Something\'s fishy');
            
            topPathRev.nodes.push(rightNode);
            topPathRev.siblings.push({ hash: leftNodeHash, idx: 0 });
            node = rightNode;
          }
        }
      }

      assert(i < leafNoBinStr.length);  // i should never go past

      var nodeKey;
      if(Acls.Helper.isRootNode(node)) {
        nodeKey = CryptoUtils.decryptKey(ownerMasterKey, node.ownerEncKey[0]);
      } else {
        nodeKey = CryptoUtils.decryptKey(ownerMasterKey, node.ownerEncKey)
      }
      //log('insertion point: ' + stri(node));
    
      var startInfo = { 
        node: node, // the node where we should start extending the path
        nodeKey: nodeKey, // this node is never a root node, so we're good decrypting here.
        childIdx: takeLeft ? 0 : 1, // the path should start in the left/0 or right/1 subtree
        leafNo: leafNoBinStr, // the leaf no. string that guided us towards the insertion node
        idx: i, // we use leafNo[i...length - 1] to guide the creation of the extended path
        topPathRev: topPathRev // the path verified from the root (reversed, from the root node down)
      };
      //log('Found insertion point: ' + stri(startInfo));
      return startInfo;
    },

    createPath: function(tree, startInfo, ownerMasterKey, theirUsername) {
      var log = function(msg) { console.log('[createPath] ' + msg); };
      // Case 2.2. There's enough space in the ACL tree, need to create our
      // own path to the new user.
      //  - involves creating new nodes along an empty or non-empty path to the root
      //    * so creating nodes starting at a non-empty node, like the root or any of its children
      //    * no = newUsers's number = acl.numUsers
      //      - Example: 11 = 1011b (from root: right, left, right, right)
      //    * I can take 'no' and traverse down from the root according to it until i find an
      //      inexistent node, then i just create the rest with the proper encrypted keys
      //    * Or, I can take 'no-1' and go up the tree. If the current node is a right child, 
      //      I keep going up. If it's a left child, then that's where I start creating my path.
      //      (I can keep track of the bits in 'no' that i need to use to 'direct' the created path)
      //    * IMPORTANT: once I find the node where to start the new path, i need to verify that it's
      //      properly linked to the root via the merkle hashes. Otherwise the server could trick me
      //      into adding the new user to another acl tree.
      var startNode = startInfo.node;
      var startDir = startInfo.childIdx;
      var topPathRev = startInfo.topPathRev;
      var leafNo = startInfo.leafNo;

      var i = startInfo.idx;
      var nodesToInsert = leafNo.length - 1 - i;
      var rightNodeKey = startInfo.nodeKey;

      //log('rightNodeKey: ' + CryptoUtils.serializeSymKey(rightNodeKey));

      if(nodesToInsert > 0) {
        // Should always be inserting the first node to the right
        assert(startDir == 1);
        var leftHash = Acls.Helper.getNode(startNode.leftChild).hash;
        var rightHash = Acls.emptyHash;
        if(startNode.hash != Acls.Helper.getNodeHashWithChildren(startNode, leftHash, rightHash)) {
          throw new Meteor.Error('path-integrity', 'The ACL path we are extending failed '
            + 'the integrity check. Something\'s fishy.');
        }

        // The root node already has the keys encrypted under rightNodeKey from extendTree
        if(!Acls.Helper.isRootNode(startNode)) {
          var startNodeKey = CryptoUtils.decryptKey(ownerMasterKey, startNode.ownerEncKey);
          assert(startNodeKey);
          assert(rightNodeKey);
          startNode.encKeys[startDir] = CryptoUtils.encryptKey(rightNodeKey, startNodeKey);
        }

        //log('Branching right from node: ' + startNode._id);
        var rightNode = {
          ownerEncKey: CryptoUtils.encryptKey(ownerMasterKey, rightNodeKey),
          encKeys: [ Acls.noKey, Acls.noKey ],      // if there are more nodes, will update
          hash: null,         // updatePathHashes will fill this in
          parent: startNode._id,
          leftChild: null,    // if there are more nodes, will update
          rightChild: null,   // no right child
        };

        // Start node's already pushed, we need to push its new right child and that 
        // child's sibling's hash
        topPathRev.nodes.push(rightNode);
        topPathRev.siblings.push({ hash: leftHash, idx: 0 });

        nodesToInsert--;

        // Build nodes, generate keys and encrypt along the path
        var lastNodeId = AclNodes.insert(rightNode);
        rightNode._id = lastNodeId;
        //log('Inserted right node: ' + stri(rightNode));
        var lastNode = rightNode;
        var lastNodeKey = rightNodeKey;
        
        // We also need to update the start node's right child
        startNode.rightChild = lastNodeId;
        assert(startNode._id);
        //AclNodes.update(startNode._id, startNode);
        AclNodes.update(startNode._id, {$set: {ownerEncKey: startNode.ownerEncKey, encKeys: startNode.encKeys, leftChild: startNode.leftChild, rightChild: startNode.rightChild, parent: startNode.parent, hash: startNode.hash}});

        while(nodesToInsert > 0) {
          var leftKey = CryptoUtils.genSymKey();
          var leftNode = {
            ownerEncKey: CryptoUtils.encryptKey(ownerMasterKey, leftKey),
            encKeys: [ Acls.noKey, Acls.noKey ], // will update in later iterations
            hash: null, // updateHashes will fill this in
            parent: lastNodeId,
            leftChild: null,
            rightChild: null,
          }

          lastNode.encKeys[0] = CryptoUtils.encryptKey(leftKey, lastNodeKey);
          
          // Insert child, update parent's child pointer
          var leftChildId = AclNodes.insert(leftNode);
          leftNode._id = leftChildId;
          lastNode.leftChild = leftChildId;

          // Update parent in DB
          assert(lastNodeId);
          //AclNodes.update(lastNodeId, lastNode);
          AclNodes.update(lastNodeId, {$set: {ownerEncKey: lastNode.ownerEncKey, encKeys: lastNode.encKeys, leftChild: lastNode.leftChild, rightChild: lastNode.rightChild, parent: lastNode.parent, hash: lastNode.hash}});
          //log('Inserted left node: ' + stri(leftNode));

          nodesToInsert--;

          lastNode = leftNode;
          lastNodeId = leftChildId;
          lastNodeKey = leftKey;

          // Push the left child we just created into our path
          topPathRev.nodes.push(lastNode);
          topPathRev.siblings.push({ hash: Acls.emptyHash, idx: 1 }); // all node's right subtrees are empty
        }
      
        // Here the AclIndex will be the left child of the leaf
        var newLeftAcl = {
          treeId: tree._id,
          nodeId: lastNodeId,
          isDeleted: false,
          keyIdx: 0,
          username: theirUsername,
        };

        newLeftAcl._id = AclIndexes.insert(newLeftAcl);

        topPathRev.nodes.push(newLeftAcl);
        topPathRev.siblings.push({ hash: Acls.emptyHash, idx: 1 });
      } else {
        // This AclIndex could be the first ACL inserted in the ACL tree, which means it's the
        // root's left child, or it could be a later ACL, which would mean it always goes to
        // the right (because a path was created for it and there's one on the left already)
        var leftAcl = AclIndexes.findOne({ nodeId: startNode._id });
        var leftAclHash = Acls.Helper.getIndexHash(leftAcl);
        //log('leftAclHash: ' + leftAclHash);

        if(startNode.hash != Acls.Helper.getNodeHashWithChildren(startNode, leftAclHash, Acls.emptyHash)) {
          throw new Meteor.Error('path-integrity', 'The last ACL tree node '
            + 'integrity check failed. Something\'s fishy.');
        }

        var leftAclEmpty = (leftAclHash == Acls.emptyHash);
        var newAcl = {
          treeId: tree._id,
          nodeId: startNode._id,
          isDeleted: false,
          keyIdx: leftAclEmpty ? 0 : 1,
          username: theirUsername,
        };

        newAcl._id = AclIndexes.insert(newAcl);

        topPathRev.nodes.push(newAcl);
        topPathRev.siblings.push({ hash: leftAclHash, idx: leftAclEmpty ? 1 : 0 });
      }

      return {
        nodes: topPathRev.nodes.reverse(),
        siblings: topPathRev.siblings.reverse(),
      };
    },

    verifyTree: function(rootNode, parentNode, depth) {
      if(Acls.debug == false)
        return;

      if(rootNode.parent != parentNode) {
        throw new Meteor.Error('Invalid parent pointer for node ID ' + rootNode._id 
          + ' at depth ' + depth + '. Expected ' + parentNode + ' but got ' + rootNode.parent);
      }

      if(rootNode.leftChild != null) {
        assert.notEqual(rootNode.leftChild, rootNode.rightChild);

        var leftChild = Acls.Helper.getNode(rootNode.leftChild);
        Acls.Helper.verifyTree(leftChild, rootNode._id, depth + 1);

        if(rootNode.rightChild != null) {
          var rightChild = Acls.Helper.getNode(rootNode.rightChild);
          Acls.Helper.verifyTree(rightChild, rootNode._id, depth + 1);
        }
      } else {
        // Can't have a rightChild when you don't have a leftChild
        assert(rootNode.rightChild == null);
      }
    },
  },
}
