Album = {
    createAlbum: function(username, masterKey, privateKey, albumName) {
        // Creates album for user username, taking care of ACL, Merkle tree, etc.
        // Also adds the album ID to the list of album Ids of this user.
        // Returns objid of created album. Not sure if even this is useful.

        var aclVer = 1;
        var timeStamp = new Date();

        var newAlbum = {
            objid: null,
            name: null,                     // Encrypted with content key for this album.
            owner: username,
            aclVer: aclVer,
            timeStamp: timeStamp,           // Last modified time.
            photos: [],                     // Should photos be stored in a separate collection instead of in a list?
            merkletreeId: null,
            signature: null            
        };

        var albumId = Albums.insert(newAlbum);
        
        var albumKey = CryptoUtils.genSymKey();
        var acl = Acls.createAcl(albumId, username, masterKey, privateKey, albumKey);

        var merkletreeId = MerkleTree.newTree(albumId, username, privateKey);

        var name = CryptoUtils.encrypt(albumKey, albumName);
        var numPhotos = 0;

        var signature = Album.Helper.signAlbum(privateKey, albumId, name, username, aclVer, timeStamp, merkletreeId, numPhotos);

        Albums.update(albumId, {$set: {objid: albumId, name: name, merkletreeId: merkletreeId, signature: signature}});
        
        Meteor.users.update(Utils.getUserId(username), {$push: { 'profile.albumIds': albumId }});

        var friends = Friends.getFriends(username);
        var trustedHash = acl.hash;
        for (i = 0; i < friends.length; i++)
            trustedHash = Acls.addToAcl(albumId, username, privateKey, masterKey, friends[i], trustedHash);

        return albumId;
    },

    deleteAlbum: function(albumId, username) {
        // Deletes specified album.
        // Deletes the entry from albums and the albumId from the user listing. No methods to delete Merkle trees or ACLs, so leaving those alone for now.
        // Also, if we ever shift to storing photos in the filesystem, those will have to be deleted.
        // Returns nothing.
        
        var album = Albums.findOne(albumId);
        if (album === null || typeof(album) === 'undefined')
            throw 'No such album.';
        
        if (Meteor.isClient)
            if (Utils.me() !== album.owner)
                throw 'Trying to delete someone else\'s album.';        // Check not really necessary (and doesn't make much sense), as this is validated against when checking permissions on database operations.
                                                                        // We don't check anything on the server, as protecting against denial of service is outside our model.
        Albums.remove(albumId);
        
        Meteor.users.update(Utils.getUserId(username), {$pull: {'profile.albumIds': albumId}});
    },

    insertPhoto: function(albumId, username, masterKey, privateKey, photo, thumbnail, trustedMerkleHash) {
        // Encrypts photo and inserts it into the album with objid albumId. username, masterKey and privateKey are of the owner of the album.
        // Takes care of encrypting photo, updating Merkle tree, etc.
        // trustedMerkleHash, optional, is the last remembered root hash of the Merkle tree of the album.
        // Returns the objid that gets assigned to the photo in the process and the new root hash of the Merkle tree.
        
        var album = Albums.findOne(albumId);
        if (album === null || typeof(album) === 'undefined')
            throw 'No such album.';
        
        if (!Album.Helper.verifySignAlbum(privateKey, album))
            throw 'Invalid signature on album.';

        var currentAcl = Acls.getAcl(albumId, username, privateKey);
        var albumKeys = Acls.getContentKeys(albumId, username, privateKey, username, masterKey, currentAcl.hash);
        var aclVer = currentAcl.acl.aclVer;
            
        var encPhoto = Album.Helper.encryptPhoto(albumKeys[aclVer-1], photo);
        var encThumbnail = Album.Helper.encryptPhoto(albumKeys[aclVer-1], thumbnail);

        var timeStamp = new Date();
        var photoEntry = {objid: (new Mongo.ObjectID())._str, aclVer: aclVer, timeStamp: timeStamp, content: encPhoto, thumbnail: encThumbnail};

        Albums.update(albumId, {$push: {photos: photoEntry}});

        var photoHash = Album.Helper.hashPhoto(photoEntry);
        var newTrustedHash = MerkleTree.insert(album.merkletreeId, photoHash, privateKey, trustedMerkleHash);
        var numPhotos = album.photos.length + 1;

        var signature = Album.Helper.signAlbum(privateKey, albumId, album.name, username, aclVer, timeStamp, album.merkletreeId, numPhotos);

        Albums.update(albumId, {$set: {aclVer: aclVer, timeStamp: timeStamp, signature: signature}});

        return {photoId: photoEntry.objid, trustedMerkleHash: newTrustedHash};
    },

    deletePhoto: function(albumId, username, masterKey, privateKey, photoId, trustedMerkleHash) {
        // Arguments same as for insertPhoto, except photoId is the objid of the photo to remove.
        // Returns the new root hash of the Merkle tree.
        
        var album = Albums.findOne(albumId);
        if (album === null || typeof(album) === 'undefined')
            throw 'No such album.';
        
        if (!Album.Helper.verifySignAlbum(privateKey, album))
            throw 'Invalid signature on album.';

        var photoEntry = null;
        for (i = 0; i < album.photos.length; i++)
            if (album.photos[i].objid === photoId)
                photoEntry = album.photos[i];

        if (photoEntry === null)
            throw 'Photo to delete does not exist.';

        var photoHash = Album.Helper.hashPhoto(photoEntry);
        var merklenodeId = MerkleTree.indexLookUp(album.merkletreeId, photoHash);

        Albums.update(albumId, {$pull: {photos: photoEntry}});
        var newTrustedMerkleHash = MerkleTree.delete(album.merkletreeId, merklenodeId, privateKey, trustedMerkleHash);
        var numPhotos = album.photos.length - 1;
        var timeStamp = new Date();

        var signature = Album.Helper.signAlbum(privateKey, albumId, album.name, username, album.aclVer, timeStamp, album.merkletreeId, numPhotos);

        Albums.update(albumId, {$set: {timeStamp: timeStamp, signature: signature}});
        
        return newTrustedMerkleHash;
    },

    getAlbumList: function(username, masterKey, privateKey, owner) {
        // Returns a list of {albumName: <album name>, objid: <album objid>} of decrypted album names of the albums of ownerUsername, if they are accessible by username. Throws exception otherwise.

        var user = Meteor.users.findOne({username: owner});
        if (user === null || typeof(user) === 'undefined')
            throw 'No such user.';

        var albumIds = user.profile.albumIds;

        var ownerPubKey = null;
        if (owner === username)
            ownerPubKey = privateKey;
        else
            ownerPubKey = Friends.getFriendPublicKey(username, masterKey, owner);

        var friendKey = masterKey;
        if (owner !== username)
            friendKey = Friends.getFriendKey(username, masterKey, owner);

        var albumList = [];
        var albumId = null, album = null;
        for (i = 0; i < albumIds.length; i++) {
            albumId = albumIds[i];

            var trustedHash = (Acls.getAcl(albumId, owner, ownerPubKey)).hash;
            var albumKey = Acls.getContentKeys(albumId, owner, ownerPubKey, username, friendKey, trustedHash)[0];
            
            album = Albums.findOne(albumId);
            if (album === null || typeof(album) === 'undefined')
                throw 'No such album.';
            if (!Album.Helper.verifySignAlbum(ownerPubKey, album))
                throw 'Invalid signature on album.';

            albumList.push({albumName: CryptoUtils.decrypt(albumKey, album.name), objid: album.objid});
        }

        return albumList;
    },

    getPhotos: function(albumId, username, masterKey, publicKey) {
        // albumId is objid of the album from which we need photos. username and masterKey are of the user who wants to see the photos.
        // Other things necessary, such as the content key for the album, will be figured out in this function itself.
        // Note that while getting photos from the Album.photos list, we need to make sure there is no duplication of objids.
        // Returns a list of all (decrypted) photos from given album, in the form {objid: <objid>, timeStamp, content: <photo>}. Throws exception if access is unauthorised.

        var start, stop;

        var album = Albums.findOne(albumId);
        if (album === null || typeof(album) === 'undefined')
            throw 'No such album.';

        var owner = album.owner;
        var ownerPubKey = null;
        if (owner === username)
            ownerPubKey = publicKey;
        else
            ownerPubKey = Friends.getFriendPublicKey(username, masterKey, owner);

        if (!Album.Helper.verifySignAlbum(ownerPubKey, album))
            throw 'Invalid signature on album.';
        
        if (!MerkleTree.verifyRootHash(album.merkletreeId, ownerPubKey))
            throw 'Invalid signature on album\'s Merkle tree.';

        if (!Album.Helper.noRepetitions(album.photos))
            throw 'List of photos in album has been tampered with.';

        var friendKey = masterKey;
        if (owner !== username)
            friendKey = Friends.getFriendKey(username, masterKey, owner);

        start = new Date();
        var trustedHash = (Acls.getAcl(albumId, owner, ownerPubKey)).hash;
        var albumKeys = Acls.getContentKeys(albumId, owner, ownerPubKey, username, friendKey, trustedHash);
        stop = new Date();
        console.log("ACL TIME: " + (stop - start));

        var photoList = [];

        var photoEntry = null, photoHash = null, merklenodeId = null, decPhoto = null;

        for (i = 0; i < album.photos.length; i++) {
            photoEntry = album.photos[i];

            start = new Date();
            photoHash = Album.Helper.hashPhoto(photoEntry);
            stop = new Date();
            console.log("HASH TIME: " + (stop - start));

            merklenodeId = MerkleTree.indexLookUp(album.merkletreeId, photoHash);

            
            if (!MerkleTree.verify(album.merkletreeId, merklenodeId, photoHash, false))
                throw 'Hash of photo not stored properly in Merkle tree.';

            start = new Date();
            var decPhoto = Album.Helper.decryptPhoto(albumKeys[photoEntry.aclVer-1], photoEntry.content);
            stop = new Date();
            console.log("DECRYPT TIME: " + (stop - start));

            photoList.push({albumId: albumId, objid: photoEntry.objid, timeStamp: photoEntry.timeStamp, content: decPhoto});
        }

        return photoList;
    },

    getPhoto: function(albumId, photoId, username, masterKey, publicKey) {
        // albumId is objid of the album from which we need the photo. 
        // photoId is objid of the photo.
        // username and masterKey are of the user who wants to see the photos.
        // Other things necessary, such as the content key for the album, will be figured out in this function itself.
        // Note that while getting photos from the Album.photos list, we need to make sure there is no duplication of objids.
        // Returns the (decrypted) photo from the given album, in the form {objid: <objid>, timeStamp, content: <photo>}.
        // Throws exception if access is unauthorised.
        // Returns null if photo doesn't exist.

        var start, stop;

        var album = Albums.findOne(albumId);
        if (album === null || typeof(album) === 'undefined')
            throw 'No such album.';

        var owner = album.owner;
        var ownerPubKey = null;
        if (owner === username)
            ownerPubKey = publicKey;
        else
            ownerPubKey = Friends.getFriendPublicKey(username, masterKey, owner);

        if (!Album.Helper.verifySignAlbum(ownerPubKey, album))
            throw 'Invalid signature on album.';
        
        if (!MerkleTree.verifyRootHash(album.merkletreeId, ownerPubKey))
            throw 'Invalid signature on album\'s Merkle tree.';

        if (!Album.Helper.noRepetitions(album.photos))
            throw 'List of photos in album has been tampered with.';

        var friendKey = masterKey;
        if (owner !== username)
            friendKey = Friends.getFriendKey(username, masterKey, owner);

        start = new Date();
        var trustedHash = (Acls.getAcl(albumId, owner, ownerPubKey)).hash;
        var albumKeys = Acls.getContentKeys(albumId, owner, ownerPubKey, username, friendKey, trustedHash);
        stop = new Date();
        console.log("ACL TIME: " + (stop - start));

        var photoElem = album.photos.filter(function(e) {return e.objid == photoId});

        if (photoElem.length != 1) {
            // Photo doesn't exist.
            return null;
        }

        var photoEntry = photoElem[0];

        start = new Date();
        var photoHash = Album.Helper.hashPhoto(photoEntry);
        stop = new Date();
        console.log("HASH TIME: " + (stop - start));

        var merklenodeId = MerkleTree.indexLookUp(album.merkletreeId, photoHash);

        if (!MerkleTree.verify(album.merkletreeId, merklenodeId, photoHash, false))
            throw 'Hash of photo not stored properly in Merkle tree.';

        start = new Date();
        var decPhoto = Album.Helper.decryptPhoto(albumKeys[photoEntry.aclVer-1], photoEntry.content);
        stop = new Date();
        console.log("DECRYPT TIME: " + (stop - start));

        return {albumId: albumId, objid: photoEntry.objid, timeStamp: photoEntry.timeStamp, content: decPhoto};
    },

    getThumbnails: function(albumId, username, masterKey, publicKey) {
        // albumId is objid of the album from which we need photos. username and masterKey are of the user who wants to see the photos.
        // Other things necessary, such as the content key for the album, will be figured out in this function itself.
        // Note that while getting photos from the Album.photos list, we need to make sure there is no duplication of objids.
        // Returns a list of all (decrypted) photos from given album, in the form {objid: <objid>, timeStamp, content: <photo>}. Throws exception if access is unauthorised.

        var start, stop;

        var album = Albums.findOne(albumId);
        if (album === null || typeof(album) === 'undefined')
            throw 'No such album.';

        var owner = album.owner;
        var ownerPubKey = null;
        if (owner === username)
            ownerPubKey = publicKey;
        else
            ownerPubKey = Friends.getFriendPublicKey(username, masterKey, owner);

        if (!Album.Helper.verifySignAlbum(ownerPubKey, album))
            throw 'Invalid signature on album.';
        
        if (!MerkleTree.verifyRootHash(album.merkletreeId, ownerPubKey))
            throw 'Invalid signature on album\'s Merkle tree.';

        if (!Album.Helper.noRepetitions(album.photos))
            throw 'List of photos in album has been tampered with.';

        var friendKey = masterKey;
        if (owner !== username)
            friendKey = Friends.getFriendKey(username, masterKey, owner);

        start = new Date();
        var trustedHash = (Acls.getAcl(albumId, owner, ownerPubKey)).hash;
        var albumKeys = Acls.getContentKeys(albumId, owner, ownerPubKey, username, friendKey, trustedHash);
        stop = new Date();
        console.log("ACL TIME: " + (stop - start));

        var photoList = [];

        var photoEntry = null, photoHash = null, merklenodeId = null, decThumbnail = null;

        for (i = 0; i < album.photos.length; i++) {
            photoEntry = album.photos[i];

            start = new Date();
            photoHash = Album.Helper.hashPhoto(photoEntry);
            stop = new Date();
            console.log("HASH TIME: " + (stop - start));

            merklenodeId = MerkleTree.indexLookUp(album.merkletreeId, photoHash);
            
            if (!MerkleTree.verify(album.merkletreeId, merklenodeId, photoHash, false))
                throw 'Hash of photo not stored properly in Merkle tree.';

            start = new Date();
            var decThumbnail = Album.Helper.decryptPhoto(albumKeys[photoEntry.aclVer-1], photoEntry.thumbnail);
            stop = new Date();
            console.log("DECRYPT TIME: " + (stop - start));

            photoList.push({albumId: albumId, objid: photoEntry.objid, timeStamp: photoEntry.timeStamp, thumbnail: decThumbnail});
        }

        return photoList;
    },

    Helper: {
        encryptPhoto: function(symKey, photo) {
            // Returns encrypted photo. Useful if we want to change what photo contains later.
            
            if (typeof(photo) === 'string')
                return CryptoUtils.encrypt(symKey, photo);

            return null;
        },

        decryptPhoto: function(symKey, encPhoto) {
            // Returns decrypted encPhoto.
            
            var decPhoto = CryptoUtils.decrypt(symKey, encPhoto);
            return decPhoto;
            //return sjcl.codec.utf8String.fromBits(decPhoto);
        },

        signAlbum: function(privateKey, albumId, name, username, aclVer, timeStamp, merkletreeId, numPhotos) {
            // Returns appropriate signature.

            var msg = EJSON.stringify({objid: albumId, name: name, owner: username, aclVer: aclVer, timeStamp: timeStamp, merkletreeId: merkletreeId, numPhotos: numPhotos});
            var signature = CryptoUtils.signString(privateKey, msg);

            return signature;
        },

        verifySignAlbum: function(publicKey, album) {
            // Verifies that the album is properly signed.

            var msg = EJSON.stringify({objid: album.objid, name: album.name, owner: album.owner, aclVer: album.aclVer, timeStamp: album.timeStamp, merkletreeId: album.merkletreeId, numPhotos: album.photos.length});

            return CryptoUtils.verifySig(publicKey, msg, album.signature);

        },

        hashPhoto: function(photoEntry) {
            // Generates the hash of photoEntry that is inserted into the Merkle tree.

            var start, stop;

            start = new Date();
            var photoString = EJSON.stringify(photoEntry);
            stop = new Date();
            console.log("STRINGIFY TIME: " + (stop - start));

            start = new Date();
            var photoHash = sjcl.codec.utf8String.toBits(CryptoJS.SHA256(photoString));
            stop = new Date();
            console.log("LOW HASH TIME: " + (stop - start));

            return photoHash;
        },

        noRepetitions: function(photos) {
            // Checks that there is no repetition in the objid's of the photos in photos.

            for (i = 0; i < photos.length; i++)
                for (j = 0; j < photos.length; j++)
                    if (i !== j && photos[i].objid === photos[j].objid)
                        return false;

            return true;
        }
    }

};


