HashChain = {
    newChain: function() {
        // Creates a new empty hash chain and adds it to the HashChains collection.
        // Returns the ID of the new hash chain.
        
        var newChain = {
            headNode: null
        };

        return HashChains.insert(newChain);
    },

    insert: function(hashchainId, value) {
        // Adds a HashChainNode with value value to the head of the hash chain whose ID is hashchainId.
        // value and hash are stored as BitArrays (defined in SJCL). If value is passed as string, it is first converted to BitArray.
        // sjcl.codec.utf8String.fromBits() and .toBits() can be used to convert between strings and BitArrays.
        // Returns headNode.hash, the hash in the new headNode.
        
        var hashchain = HashChains.findOne(hashchainId);
        if (hashchain === null || typeof(hashchain) === 'undefined')
            throw 'Hash chain does not exist.';

        if (typeof(value) === 'string') 
            value = sjcl.codec.utf8String.toBits(value);

        var prevNodeId = hashchain.headNode;

        var hash = sjcl.hash.sha256.hash(value);
        if (prevNodeId !== null)
        {
            var prevNode = HashChainNodes.findOne(prevNodeId);
            if (prevNode === null || typeof(prevNode) === 'undefined')
                throw 'Hash chain has invalid headNode.';

            hash = sjcl.hash.sha256.hash(value.concat(prevNode.hash));
        }
            
        var newNode = {
            value: value,
            hash: hash,
            prevNode: prevNodeId
        };
        
        var newNodeId = HashChainNodes.insert(newNode);
        HashChains.update(hashchainId, {$set: {'headNode': newNodeId}});

        return hash;
    },

    getKNodes: function(hashchainId, numberOfNodes) {
        // Return the top numberOfNodes number of entries in the hash chain, along with the hash, and also the hash of the (numberOfNodes+1)th node as: 
        // {nodes: {value: HashChainNode.value, hash: HashChainNode.hash}], nextHash: nextHash}.

        var hashchain = HashChains.findOne(hashchainId);
        if (hashchain === null || typeof(hashchain) === 'undefined')
            throw 'Hash chain does not exist.';

        var ret = [];
        var nextHash = null;

        var headNodeId = hashchain.headNode;
        if (headNodeId === null)
            return {nodes: ret, nextHash: nextHash};

        var headNode = HashChainNodes.findOne(headNodeId);
        if (headNode === null || typeof(headNode) === 'undefined')
            throw 'Hash chain has invalid headNode.';

        var node = headNode;
        for (i = 0; i < numberOfNodes && node !== null; i++) {
            ret.push({value: node.value, hash: node.hash});
            
            if (node.prevNode === null) {
                node = null;
                nextHash = null;
            }
            else {
                node = HashChainNodes.findOne(node.prevNode);
                if (node === null || typeof(node) === 'undefined')
                    throw 'Somewhere along this hash chain is an invalid prevNode.';
                nextHash = node.hash;
            }
        }

        return {nodes: ret, nextHash: nextHash};
    },

    getTop: function(hashchainId) {
        // Returns just the value and hash, and the nextHash, at the head of the chain, as {node: {value: headNode.value, hash: headNode.hash}, nextHash: nextHash}.
      
        var ret = HashChain.getKNodes(hashchainId, 1);

        if (ret.nodes.length === 0)
            return {node: null, nextHash: null};
        
        return {node: ret.nodes[0], nextHash: ret.nextHash};
    },

    deleteKNodes: function(hashchainId, numberOfNodes) {
        // Deletes the top numberOfNodes nodes in the hash chain, or all nodes if less than that many exist.
        // Returns the same thing as getKNodes. Is this redundant? There seem to be uses for both functions.
        
        var hashchain = HashChains.findOne(hashchainId);
        if (hashchain === null || typeof(hashchain) === 'undefined')
            throw 'Hash chain does not exist.';

        var ret = [];
        var nextHash = null;

        var headNodeId = hashchain.headNode;
        if (headNodeId === null)
            return {nodes: ret, nextHash: nextHash};

        var headNode = HashChainNodes.findOne(headNodeId);
        if (headNode === null || typeof(headNode) === 'undefined')
            throw 'Hash chain has invalid headNode.';

        var node = headNode;
        var newHeadNodeId = headNodeId;
        var toBeDeletedIds = [];
        for (i = 0; i < numberOfNodes && node !== null; i++) {
            ret.push({value: node.value, hash: node.hash});
            
            newHeadNodeId = node.prevNode;
            toBeDeletedIds.push(node._id);

            if (node.prevNode === null) {
                node = null;
                nextHash = null;
            }
            else {
                node = HashChainNodes.findOne(node.prevNode);
                if (node === null || typeof(node) === 'undefined')
                    throw 'Somewhere along this hash chain is an invalid prevNode.';
                nextHash = node.hash;
            }
        }

        HashChains.update(hashchainId, {$set: {'headNode': newHeadNodeId}});
        
        for (i = 0; i < toBeDeletedIds.length; i++)
            HashChainNodes.remove(toBeDeletedIds[i]);

        return {nodes: ret, nextHash: nextHash};
    },

    deleteTop: function(hashchainId) {
        // Deletes the head of the hash chain.
        // Returns the same thing as getTop.
      
        var ret = HashChain.deleteKNodes(hashchainId, 1);

        if (ret.nodes.length === 0)
            return {node: null, nextHash: null};
        
        return {node: ret.nodes[0], nextHash: ret.nextHash};
    }
}
