MerkleTree = {
    newTree: function(objid, owner, ownerPrivateKey) {
        // Creates a new empty Merkle tree and adds it to the MerkleTrees collection.
        // Also adds the initial signature to the tree.
        // Returns the ID of the new Merkle tree.
        
        var msg = EJSON.stringify({objid: objid, owner: owner, rootHash: null});
        var signature = CryptoUtils.signString(ownerPrivateKey, msg);

        var newTree = {
            objid: objid,            // objid of the object (album, etc.) for which this Merkle tree is used.
            owner: owner,           // Can add signer if any tree is ever edited by more than one person.
            rootNode: null,
            leafNodes: [],
            emptyLeafNodes: [],
            treeDepth: -1,          // Single node is depth 0.
            signature: signature    // sign(objid, rootNode.hash, leafNodes)
        };

        return MerkleTrees.insert(newTree);
    },
    
    hashTreeNodes: function(leftNode, rightNode) {
        // leftNode and rightNode are Merkle tree nodes (not their IDs).
        // Used which verifying Merkle hashes - the nodes passed are children of some other node.
        // Returns hash(leftNode.isEmpty, leftNode.hash, rightNode.isEmpty, rightNode.hash).
        
        var toBeHashed = [];

        var leftIsEmpty = leftNode.isEmpty ? [1] : [0];
        var rightIsEmpty = rightNode.isEmpty ? [1] : [0];

        toBeHashed = toBeHashed.concat(leftIsEmpty);
        toBeHashed = toBeHashed.concat(leftNode.hash);
        toBeHashed = toBeHashed.concat(rightIsEmpty);
        toBeHashed = toBeHashed.concat(rightNode.hash);

        return sjcl.hash.sha256.hash(toBeHashed);
    },

    enlargeTree: function(merkletreeId) {
        // Doubles the size of the given tree.
        // Returns updated merkletree object.

        var createEmptySubTree = function(depth, parentId) {
            // Creates an empty full binary subtree of depth depth, and creates necessary MerkleTreeNodes entries. parentId is taken to be the parent of the subtree.
            // Note that nothing in the node with ID parentId is actually modified here.
            // Returns an object like {emptyRoot: <root node of created subtree>, emptyNodes: [list of IDs of newly created leaf nodes]}.
            
            var newNode = {
                isEmpty: false,
                hash: [],
                parent: parentId,
                leftChild: null,
                rightChild: null
            };

            if (depth === 0) {
                newNode.isEmpty = true;
                var newNodeId = MerkleTreeNodes.insert(newNode);
                newNode._id = newNodeId;

                return {emptyRoot: newNode, emptyNodes: [newNodeId]};
            }

            var newNodeId = MerkleTreeNodes.insert(newNode);
            var leftCall = createEmptySubTree(depth-1, newNodeId);
            var rightCall = createEmptySubTree(depth-1, newNodeId);
            
            newNode._id = newNodeId;
            newNode.leftChild = leftCall.emptyRoot._id;
            newNode.rightChild = rightCall.emptyRoot._id;
            newNode.hash = MerkleTree.hashTreeNodes(leftCall.emptyRoot, rightCall.emptyRoot);

            MerkleTreeNodes.update(newNodeId, {$set: {leftChild: newNode.leftChild, rightChild: newNode.rightChild, hash: newNode.hash}});

            var emptyNodes = leftCall.emptyNodes.concat(rightCall.emptyNodes);

            return {emptyRoot: newNode, emptyNodes: emptyNodes};
        };

        var merkletree = MerkleTrees.findOne(merkletreeId);
        if (merkletree === null || typeof(merkletree) === 'undefined')
            throw 'Merkle tree does not exist.';
        
        var newRoot = {
            isEmpty: false,
            hash: [],
            parent: null,
            leftChild: null,
            rightChild: null
        };

        if (merkletree.rootNode === null) {
            newRoot.isEmpty = true;
            var newRootId = MerkleTreeNodes.insert(newRoot);

            merkletree.rootNode = newRootId;
            merkletree.emptyLeafNodes = [newRootId];
            merkletree.leafNodes = [newRootId];
            merkletree.treeDepth = 0;

            MerkleTrees.update(merkletreeId, {$set: {rootNode: merkletree.rootNode, emptyLeafNodes: merkletree.emptyLeafNodes, leafNodes: merkletree.leafNodes, treeDepth: merkletree.treeDepth}});

            return merkletree;
        }
        else {
            var oldRoot = MerkleTreeNodes.findOne(merkletree.rootNode);
            if (oldRoot === null || typeof(oldRoot) === 'undefined')
                throw 'Merkle tree has invalid rootNode.';
            var oldRootId = oldRoot._id;

            var newRootId = MerkleTreeNodes.insert(newRoot);

            var emptySubTree = createEmptySubTree(merkletree.treeDepth, newRootId);
            var emptyRoot = emptySubTree.emptyRoot;

            MerkleTreeNodes.update(oldRootId, {$set: {parent: newRootId}});
            MerkleTreeNodes.update(newRootId, {$set: {leftChild: oldRootId, rightChild: emptyRoot._id, hash: MerkleTree.hashTreeNodes(oldRoot, emptyRoot)}});

            merkletree.rootNode = newRootId;
            merkletree.emptyLeafNodes = merkletree.emptyLeafNodes.concat(emptySubTree.emptyNodes);
            merkletree.leafNodes = merkletree.leafNodes.concat(emptySubTree.emptyNodes);
            merkletree.treeDepth = merkletree.treeDepth + 1;
            MerkleTrees.update(merkletreeId, {$set: {rootNode: merkletree.rootNode, emptyLeafNodes: merkletree.emptyLeafNodes, leafNodes: merkletree.leafNodes, treeDepth: merkletree.treeDepth}});

            return merkletree;
        }
    },

    insertAt: function(merklenodeId, value) {
        // Sets the hash at node with ID merklenodeID to value, and propagates it up to the root of the tree it is in.
        // value is stored as a BitArray - converted from string if necessary.
        // That the node was empty and is in the correct Merkle tree is to be checked by the client before asking the server to call this function.
        // Does _not_ care about any MerkleTree.emptyLeafNodes, node.isEmpty or MerkleTreeIndex. Those need to be managed by the caller.
        // Returns the root hash of the resulting tree.
        
        var updateHash = function(parentId, oneChild) {
            // Updates the value of the hash in the node whose ID is parentId.
            // This is to be called after updating the hash in one of its children, which is the oneChild node.
            // Returns the updated hash of the root of the tree.

            var parentNode = MerkleTreeNodes.findOne(parentId);
            if (parentNode === null || typeof(parentNode) === 'undefined')
                throw 'A "parent" entry is invalid in this tree.';

            if (parentNode.leftChild === null || parentNode.rightChild === null)
                throw 'A child entry is invalid in this tree.';

            var leftChild = null, rightChild = null;
            if (parentNode.leftChild === oneChild._id) {
                leftChild = oneChild;
                rightChild = MerkleTreeNodes.findOne(parentNode.rightChild);
            }
            else 
            if (parentNode.rightChild === oneChild._id) {
                rightChild = oneChild;
                leftChild = MerkleTreeNodes.findOne(parentNode.leftChild);
            }
            else {
                throw 'Someone doesn\'t know their family well enough.';
            }

            parentNode.hash = MerkleTree.hashTreeNodes(leftChild, rightChild);
            MerkleTreeNodes.update(parentId, {$set: {hash: parentNode.hash}});

            if (parentNode.parent === null)
                return parentNode.hash;
            return updateHash(parentNode.parent, parentNode);
        };

        var merklenode = MerkleTreeNodes.findOne(merklenodeId);
        if (merklenode === null || typeof(merklenode) === 'undefined')
            throw 'Merkle node does not exist.';

        if (typeof(value) === 'string') 
            value = sjcl.codec.utf8String.toBits(value);

        merklenode.hash = value;
        MerkleTreeNodes.update(merklenodeId, {$set: {hash: merklenode.hash}});

        if (merklenode.parent === null)
            return merklenode.hash;
        return updateHash(merklenode.parent, merklenode);
    },

    insert: function(merkletreeId, value, ownerKeyPair, trustedHash) {
        // Inserts value value into the tree with ID merkletreeId, enlarging the tree if needed. value is stored as a BitArray - converted from string if necessary.
        // Adds (merkletree, value, node) to MerkleTreeLeafIndex.    
        // Perform checks on whether nodes in the MerkleTree.emptyLeafNodes are actually empty, and the root hash is correct.
        // trustedHash is an optional argument that, if given, ensures that the root hash on the tree before inserting is trustedHash.
        // Returns the root hash of the resulting tree.    

        var merkletree = MerkleTrees.findOne(merkletreeId);
        if (merkletree === null || typeof(merkletree) === 'undefined')
            throw 'Merkle tree does not exist.';

        /*
        var msg = null, oldRootHash = null;
        if (merkletree.rootNode === null) 
            msg = EJSON.stringify({objid: merkletree.objid, owner: merkletree.owner, rootHash: null});
        else {
            var rootNode = MerkleTreeNodes.findOne(merkletree.rootNode);
            if (rootNode === null || typeof(rootNode) === 'undefined')
                throw 'Merkle tree has invalid rootNode.';
            msg = EJSON.stringify({objid: merkletree.objid, owner: merkletree.owner, rootHash: rootNode.hash});
            oldRootHash = rootNode.hash;
        }
            
        if (!CryptoUtils.verifySig(ownerKeyPair, msg, merkletree.signature))
            throw 'Invalid signature on Merkle tree.';

        if (trustedHash !== null && typeof(trustedHash) !== 'undefined' && !sjcl.bitArray.equal(trustedHash, oldRootHash))
            throw 'Merkle tree root hash (' + oldRootHash + ') does not match last remembered hash (' + trustedHash + ').';
        */
        if (!MerkleTree.verifyRootHash(merkletreeId, ownerKeyPair, trustedHash))
            throw 'Invalid signature on Merkle tree.';

        if (typeof(value) === 'string') 
            value = sjcl.codec.utf8String.toBits(value);

        if (merkletree.emptyLeafNodes.length === 0) 
            merkletree = MerkleTree.enlargeTree(merkletreeId);

        var emptyNodeId = merkletree.emptyLeafNodes[0];

        if (!MerkleTree.verify(merkletreeId, emptyNodeId, '', true))
            throw 'Verification at purported empty leaf failed.';

        MerkleTreeNodes.update(emptyNodeId, {$set: {isEmpty: false}});
        var newRootHash = MerkleTree.insertAt(emptyNodeId, value);
        
        MerkleTrees.update(merkletreeId, {$pop: {emptyLeafNodes: -1}});
        MerkleTreeLeafIndex.insert({treeId: merkletreeId, value: value, nodeId: emptyNodeId});


        merkletree = MerkleTrees.findOne(merkletreeId);
        msg = EJSON.stringify({objid: merkletree.objid, owner: merkletree.owner, rootHash: newRootHash});
        var signature = CryptoUtils.signString(ownerKeyPair, msg);
        MerkleTrees.update(merkletreeId, {$set: {signature: signature}});


        return newRootHash;
        //return emptyNodeId;
    },

    delete: function(merkletreeId, merklenodeId, ownerKeyPair, trustedHash) {
        // Removes the node with ID merklenodeId from the Merkle tree with ID merkletreeId.
        // Deletes (merkletree, *, merkletreenode) from MerkleTreeLeafIndex.
        // Performs the same checks as insert.
        // Returns the root hash of the resulting tree.

        var merkletree = MerkleTrees.findOne(merkletreeId);
        if (merkletree === null || typeof(merkletree) === 'undefined')
            throw 'Merkle tree does not exist.';

        /*
        var msg = null, oldRootHash = null;
        if (merkletree.rootNode === null) 
            msg = EJSON.stringify({objid: merkletree.objid, owner: merkletree.owner, rootHash: null});
        else {
            var rootNode = MerkleTreeNodes.findOne(merkletree.rootNode);
            if (rootNode === null || typeof(rootNode) === 'undefined')
                throw 'Merkle tree has invalid rootNode.';
            msg = EJSON.stringify({objid: merkletree.objid, owner: merkletree.owner, rootHash: rootNode.hash});
            oldRootHash = rootNode.hash;
        }
            
        if (!CryptoUtils.verifySig(ownerKeyPair, msg, merkletree.signature))
            throw 'Invalid signature on Merkle tree.';

        if (trustedHash !== null && typeof(trustedHash) !== 'undefined' && !sjcl.bitArray.equal(trustedHash, oldRootHash))
            throw 'Merkle tree root hash (' + oldRootHash + ') does not match last remembered hash (' + trustedHash + ').';
        */
        
        if (!MerkleTree.verifyRootHash(merkletreeId, ownerKeyPair, trustedHash))
            throw 'Invalid signature on Merkle tree.';

        if (!MerkleTree.verify(merkletreeId, merklenodeId, null, false))
            throw 'Verification at node to be deleted failed. Either it is empty, or hashes are wrong.';

        MerkleTreeNodes.update(merklenodeId, {$set: {isEmpty: true}});
        var newRootHash = MerkleTree.insertAt(merklenodeId, '');              // Delete by setting value at node to empty.

        MerkleTrees.update(merkletreeId, {$addToSet: {emptyLeafNodes: merklenodeId}});
        var merkleIndexId = MerkleTreeLeafIndex.findOne({treeId: merkletreeId, nodeId: merklenodeId})._id;
        MerkleTreeLeafIndex.remove(merkleIndexId);
        

        merkletree = MerkleTrees.findOne(merkletreeId);
        msg = EJSON.stringify({objid: merkletree.objid, owner: merkletree.owner, rootHash: newRootHash});
        var signature = CryptoUtils.signString(ownerKeyPair, msg);
        MerkleTrees.update(merkletreeId, {$set: {signature: signature}});


        return newRootHash;
    },

    verifyRootHash: function(merkletreeId, ownerPubKey, trustedHash) {
        // Checks the signature on the Merkle tree.
        // If trustedHash is not null, checks that the root hash is equal to it.
        
        var merkletree = MerkleTrees.findOne(merkletreeId);
        if (merkletree === null || typeof(merkletree) === 'undefined')
            throw 'Merkle tree does not exist.';

        var msg = null, oldRootHash = null;
        if (merkletree.rootNode === null) 
            msg = EJSON.stringify({objid: merkletree.objid, owner: merkletree.owner, rootHash: null});
        else {
            var rootNode = MerkleTreeNodes.findOne(merkletree.rootNode);
            if (rootNode === null || typeof(rootNode) === 'undefined')
                throw 'Merkle tree has invalid rootNode.';
            msg = EJSON.stringify({objid: merkletree.objid, owner: merkletree.owner, rootHash: rootNode.hash});
            oldRootHash = rootNode.hash;
        }
            
        if (!CryptoUtils.verifySig(ownerPubKey, msg, merkletree.signature))
            return false;

        if (trustedHash !== null && typeof(trustedHash) !== 'undefined' && !sjcl.bitArray.equal(trustedHash, oldRootHash))
            throw 'Merkle tree root hash (' + oldRootHash + ') does not match last remembered hash (' + trustedHash + ').';

        return true;
    },

    verify: function(merkletreeId, merklenodeId, value, isEmpty) {
        // Checks that merklenode.hash=value, that all the hashes from this node up to the root check out, and that the root is the rootNode of merkletree.
        // merklenodeId, merkletreeId could just be the tree or node itself. This is convenient.
        // Can pass value/isEmpty = null to just test hashes and not the value.
        // Returns True or False.
        

        var verifyUp = function(parentId, oneChild) {
            // Verifies the hash at node with ID parentId, and recursively calls itself on the node's parent.
            // oneChild is a child (node, not parent) of this node.
            // Returns true if all the checks pass, false otherwise.

            var parentNode = MerkleTreeNodes.findOne(parentId);
            if (parentNode === null || typeof(parentNode) === 'undefined')
                throw 'A "parent" entry is invalid in this tree.';

            if (parentNode.leftChild === null || parentNode.rightChild === null)
                throw 'A child entry is invalid in this tree.';

            var leftChild = null, rightChild = null;
            if (parentNode.leftChild === oneChild._id) {
                leftChild = oneChild;
                rightChild = MerkleTreeNodes.findOne(parentNode.rightChild);
            }
            else 
            if (parentNode.rightChild === oneChild._id) {
                rightChild = oneChild;
                leftChild = MerkleTreeNodes.findOne(parentNode.leftChild);
            }
            else {
                throw 'Someone doesn\'t know their family well enough.';
            }

            if (!sjcl.bitArray.equal(parentNode.hash, MerkleTree.hashTreeNodes(leftChild, rightChild)))
                return false;

            if (parentNode.parent === null) {
                if (merkletree.rootNode === parentId)
                    return true;
                else
                    return false;
            }
            return verifyUp(parentNode.parent, parentNode);
        };
        
        var merkletree = null;
        if (typeof(merkletreeId) === 'string')
            merkletree = MerkleTrees.findOne(merkletreeId);
        else 
            merkletree = merkletreeId;
        if (merkletree === null || typeof(merkletree) === 'undefined')
            throw 'Merkle tree does not exist.';

        var merklenode = null;
        if (typeof(merklenodeId) === 'string')
            merklenode = MerkleTreeNodes.findOne(merklenodeId);
        else
            merklenode = merklenodeId;
        if (merklenode === null || typeof(merklenode) === 'undefined')
            throw 'Merkle node does not exist.';

        if (typeof(value) === 'string') 
            value = sjcl.codec.utf8String.toBits(value);

        if (value !== null && !sjcl.bitArray.equal(merklenode.hash, value)) 
            return false;

        if (isEmpty !== null && merklenode.isEmpty !== isEmpty)
            return false;
        
        if (merklenode.parent === null) {
            if (merkletree.rootNode === merklenodeId)
                return true;
            else
                return false;
        }

        return verifyUp(merklenode.parent, merklenode);
    },

    indexLookUp: function(merkletreeId, value) {
        // Returns the Id of the node in which value is stored in merkletree with id merkletreeId.
        // FIXME: Should this check the validity of the MerkleTreeLeafIndex entry? Perhaps.
        // Currently it verifies that the value stored is consistent with the rest of the tree.
        
        if (typeof(value) === 'string') 
            value = sjcl.codec.utf8String.toBits(value);

        var entry = MerkleTreeLeafIndex.findOne({treeId: merkletreeId, value: value});

        if (entry === null || typeof(entry) === 'undefined')
            return null;

        var merklenode = MerkleTreeNodes.findOne(entry.nodeId);
        if (merklenode === null || typeof(merklenode) === 'undefined')
            throw 'Merkle node does not exist.';
        if (!sjcl.bitArray.equal(merklenode.hash, value))
            throw 'Corrupt entry in MerkleTreeNodeIndex: ' + EJSON.stringify(entry) + 'referring to node: ' + EJSON.stringify(merklenode) + ', but value was supposed to be: ' + value + '.';

        return entry.nodeId;
    }
}
