/**
 * CryptoUtils: utility functions for crypto
 * 
 * Derivative of Mylar's 'basic-crypto' package
 *
 * NOTE: we use the following types in the documentation
 * - String - a normal JavaScript string
 * - HexStr - a String that stores hex-encoded data
 * - SymmetricKey - an array of 4-byte words
 * - KeyPair - an asymmetric KeyPair (right now this is an ellipticjs type)
 * - CipherText - a ciphertext object (right now this is an SJCL type)
 * - Mac - a message authentication code (MAC) object (right now this is an SJCL type)
 */

CryptoUtils = (function () {
  var EC = ellipticjs.ec;
  var ec = new EC('ed25519');

  return {
    // @param minInclusive the min as an unsigned integer 
    // @param maxExclusive the max as an unsigned integer
    //
    // @return a random number in [minInclusive, maxExclusive) range
    randomInt: function(minInclusive, maxExclusive) {
      // The triple > is for unsigned right shifting the value and converting it into
      // an unsigned number
      // TODO: SECURITY: set paranoia parameter
      return ((sjcl.random.randomWords(1)[0] >>> 0) % (maxExclusive - minInclusive)) + minInclusive;
    },

    randomBytes: function(num) {
      if(!num) {
        num = 8;
      } else if(num > 32) {
        throw new Meteor.Error('SJCL only supports up to 32-byte keys');
      } else {
        num = num/4; // SJCL generates 4 byte words
      }
      // TODO: SECURITY: set paranoia parameter
      return sjcl.codec.hex.fromBits(sjcl.random.randomWords(num));
    },

    // @return a 256-bit/32-byte random HexStr
    randomSalt: function() {
      // TODO: SECURITY: set paranoia parameter
      return sjcl.codec.hex.fromBits(sjcl.random.randomWords(8));
    },

    //// @return returns an all zeros symmetric key
    //getZeroSymKey: function() {
    //  return sjcl.codec.hex.toBits("0000000000000000000000000000000000000000000000000000000000000000");
    //},

    // @return an ED25519 KeyPair object that can be used for signing/verifying
    genAsymKeyPair: function() {
      //return new ellipticjs.ec('ed25519').genKeyPair();
      return ec.genKeyPair();
    },

    // @param keyPair the public key to serialize (as a KeyPair object)
    //
    // @return the hex-encoded public key as a HexStr
    serializePublicKey: function(keyPair) {
      return keyPair.getPublic(null, 'hex');
    },

    // @param publicKeyHexStr the public key as a HexStr
    //
    // @return the public key as a KeyPair object (no private key included)
    unserializePublicKey: function(publicKeyHexStr) {
      return ec.keyPair(publicKeyHexStr, 'hex');
    },

    // @param keyPair the private key to serialize (as a KeyPair object)
    //
    // @return the hex-encoded private key as a HexStr
    serializePrivateKey: function(keyPair) {
      return keyPair.getPrivate('hex');
    },

    // @param privateKeyHexStr the private key as a HexStr
    //
    // @return the private key as a KeyPair object (the public key will be included as well)
    unserializePrivateKey: function(privateKeyHexStr) {
      return ec.keyPair({ priv: privateKeyHexStr, pub: null });
    },

    // @param keyPair the public-private key pair (only needs the public part to be set)
    //  - this can be either a KeyPair object
    //  - or it can be a HexStr public key (obtained from CryptoUtils.serializePublicKey)
    //
    // @return returns the fingerprint of the key as a HexStr. the fingerprint is just
    // the SHA256 hash over the serialized public key string
    getPublicKeyFingerprint: function(keyPair) {
      var serialized = keyPair;

      if(typeof keyPair != 'string')
        serialized = CryptoUtils.serializePublicKey(keyPair);

      return sjcl.codec.hex.fromBits(sjcl.hash.sha256.hash(serialized));
    },

    // pretty self explanatory, see above too.
    verifyFingerprint: function(keyPair, fp) {
        return fp == CryptoUtils.getPublicKeyFingerprint(keyPair);
    },

    // @param secretKey the secret key (as a KeyPair object) used to sign
    // @param str the String object to sign. must be a String.
    //
    // @return the signature as a DER-encoded HexStr
    signString: function(secretKey, str) {
      //var msg = ellipticjs.utils.toArray(str);
      var msg = sjcl.hash.sha256.hash(str);             // Need to do this because is toArray() is used, the array returned is truncated to certain length before signing.

      if(typeof secretKey == 'string') {
        throw new Meteor.Error('bad-key', 'Secret key must be a proper KeyPair object, use unserializePrivateKey() on it. You provided \'' + stri(secretKey) + '\'');
      }

      return secretKey.sign(msg).toDER('hex');
    },

    // @param publicKey the public key (as a KeyPair object) used to verify the signature with
    // @param msgStr the message String whose signature we are verifying (a String)
    // @param sigHexStr the DER-encoded signature as a HexStr
    //
    // @return true if signature(msgStr) == sigHexStr, false otherwise
    verifySig: function(publicKey, msgStr, sigHexStr) {
      check(msgStr, String);
      check(sigHexStr, String);
      if(typeof publicKey == 'string') {
        throw 'public key must be a proper KeyPair object, use unserializePublicKey() on it';
      }

      //return publicKey.verify(ellipticjs.utils.toArray(msgStr), sigHexStr);
      return publicKey.verify(sjcl.hash.sha256.hash(msgStr), sigHexStr);
    },

    // @param key the SymmetricKey to MAC the data with
    // @param str the message as a String that will be MAC'd
    //
    // @param str a HexStr with the MAC of the str parameter under the specified key
    mac: function(key, str) {
      var hmac = new sjcl.misc.hmac(key, sjcl.hash.sha256);
      return sjcl.codec.hex.fromBits(hmac.mac(str));
    },

    verifyMac: function(key, str, mac) {
      var actualMac = CryptoUtils.mac(key, str);
      return actualMac == mac;
    },

    hash: function(str) {
      var h = sjcl.codec.hex.fromBits(sjcl.hash.sha256.hash(str));
      //console.log('[Crypto.hash] Hashing \'' + str + '\' -> ' + h);
      return h;
    },

    //asymEncrypt: function(publicKey, str) {
    //  var msg = ellipticjs.utils.toArray(str);
    //  
    //  if(typeof publicKey == 'string') {
    //    throw 'public key must be a proper KeyPair object, use unserializePublicKey() on it';
    //  }

    //  throw 'not implemented, because ellipticjs only supports signing';
    //},

    //asymDecrypt: function(secretKey, ctxtStrHex) {
    //  throw 'not implemented, because ellipticjs only supports signing';
    //},

    // @param secret the secret TODO: type?
    // @param salt the salt TODO: type, hexadecimal string works
    //
    // @return a SymmetricKey
    keyDerivationFunction: function(secret, salt) {
      return sjcl.misc.pbkdf2(secret, salt, 10000, 256);
    },

    // @param myPrivateKey a KeyPair object with your private key (a KeyPair)
    // @param theirPublickey a KeyPair object with the other party's public key (a KeyPair)
    //
    // @return a 32-byte hex-encoded symmetric key as a HexStr
    ecdhDeriveKey: function(myPrivateKey, theirPublicKey) {
      if(typeof theirPublicKey == 'string') {
        throw 'public key must be a proper KeyPair object, use unserializePublicKey() on it';
      }
      if(typeof myPrivateKey == 'string') {
        throw 'secret key must be a proper KeyPair object, use unserializePrivateKey() on it';
      }

      return myPrivateKey.derive(theirPublicKey.getPublic()).toString(16);
    },

    // @return a 32-byte/256-bit SymmetricKey. the returned key 
    // can be used in CryptoUtils.encrypt/decrypt functions.
    genSymKey: function () {
      // sjcl's max size for key is 8 random words, which is weird
      // TODO: SECURITY: set paranoia parameter
      var symKey = sjcl.random.randomWords(8);
      return symKey;
    },

    // @param key a SymmetricKey obtained via CryptoUtils functions
    serializeSymKey: function (key) {
      //return EJSON.stringify(key, { canonical: true });
      return sjcl.codec.hex.fromBits(key);
    },

    // @param ct a CipherText object obtained from CryptoUtils functions
    serializeCipherText: function (ct) {
      return EJSON.stringify(ct, { canonical: true });
    },

    // @param pk the SymmetricKey used to encrypt with, obtained from
    //  - sjcl.misc.pbkdf2
    //  - CrypotUtils.keyDerivationFunction
    //  - CryptoUtils.genSymKey
    // @param data the data to be encrypted (a String, or an SJCL array)
    //
    // @return a serializable CipherText object containing the authenticated ciphertext
    encrypt: function (pk, data) {
        return sjcl.encrypt(pk, data);
    },

    // @param pk the SymmetricKey used to encrypt with.
    // @param key the SymmetricKey that will be encrypted
    //
    // @return converts the key that is being encrypted to hexadecimal, encrypts
    // the key and returns a serializable CipherText object
    encryptKey: function (pk, key) {
      return CryptoUtils.encrypt(pk, sjcl.codec.hex.fromBits(key));
    },

    // @param pk the SymmetricKey used to decrypt with. Can be obtained from:
    //  - sjcl.misc.pbkdf2
    //  - CrypotUtils.keyDerivationFunction
    //  - CryptoUtils.genSymKey
    // @param ct the data to be decrypted (must be a CipherText object w/ IV, tag, etc.)
    //
    // @return the plaintext as a String. throws an exception if the ct CipherText has
    // been tampered with (integrity check)
    decrypt: function (sk, ct) {
        assert(sk, 'empty secret key');
        return sjcl.decrypt(sk, ct);
    },

    // @param sk the SymmetricKey used to decrypt with.
    // @param key the encrypted SymmetricKey as a CipherText object.
    //
    // @return decrypts the key and returns the key as a SymmetricKey object, so that it can
    // be used in other CryptoUtils calls. throws an exception if the CipherText object has
    // been tampered with (integrity check)
    decryptKey: function (sk, key) {
      return sjcl.codec.hex.toBits(CryptoUtils.decrypt(sk, key));
    },

    // authenticated encryption
    //authEncrypt: function (sk, data, adata) {
    //    if (data === undefined) {
    //        console.log('sym_encrypt: undefined', data);
    //        return;
    //    }

    //    var prp = new sjcl.cipher['aes'](sk);
    //    var iv = sjcl.random.randomWords(4, 0);
    //    var pt = sjcl.codec.utf8String.toBits(data);
    //    var ct = sjcl.mode['ccm'].encrypt(prp, pt, iv, adata);
    //    // Undefined adata/tag; might be useful later?

    //    return sjcl.codec.base64.fromBits(iv.concat(ct), 1);
    //},

    //authDecrypt: function (sk, data, adata) {
    //    if (data === undefined) {
    //        console.log('sym_decrypt: undefined', data);
    //        return;
    //    }

    //    var prp = new sjcl.cipher['aes'](sk);
    //    var bits = sjcl.codec.base64.toBits(data);
    //    var iv = bits.slice(0, 4);
    //    var ct = bits.slice(4);
    //    var pt = sjcl.mode['ccm'].decrypt(prp, ct, iv, adata);
    //    return sjcl.codec.utf8String.fromBits(pt);
    //},

    ///* Starting with a secret key sk,
    // unwraps keys in chain, until it obtains
    // the secret key at the end of the chain. */
    //chain_decrypt: function (chain, keys) {
    //    _.each(chain, function (wk) {
    //        var unwrapped;
    //        if (wk.isSym == undefined || wk.isSym == null) {
    //            throw new Error("invalid wrapped key");
    //        }
    //        if (wk.isSym) {
    //            unwrapped = CryptoUtils.sym_decrypt(keys.sym_key, wk.wkey);
    //        } else {
    //            unwrapped = CryptoUtils.decrypt(keys.decrypt, wk.wkey);
    //        }
    //        keys = deserialize_keys(unwrapped);
    //    });
    //    return keys;
    //},

    //chain_verify: function (chain) {
    //    chain = _.map(chain, function (cert) {
    //        var hash = sjcl.hash.sha256.hash(cert.m);
    //        var pk = cert.pk;
    //        try {
    //            pk.verify(hash, cert.sig);
    //            return true;
    //        } catch (e) {
    //            return false;
    //        }
    //    });
    //    return _.every(chain, _.identity);
    //}
  };
})();
